# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset class for Placesfull (256x256) dataset."""
import csv
import os
import urllib

from etils import epath
import tensorflow_datasets.public_api as tfds

_BASE_URL = "http://data.csail.mit.edu/places/places365/"
_TRAIN_URL = "Images256.tar"
_FILE_ANNOTATION_URL = "filelist_placesfull.tar"

_IMAGE_SHAPE = (256, 256, 3)

_LABELS_FNAME = "datasets/placesfull/categories.txt"


class Builder(tfds.core.GeneratorBasedBuilder):
  """Placesfull Images dataset."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    names_file = tfds.core.tfds_path(_LABELS_FNAME)
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(shape=_IMAGE_SHAPE),
            "label": tfds.features.ClassLabel(names_file=names_file),
            "filename": tfds.features.Text(),
        }),
        supervised_keys=("image", "label", "filename"),
        homepage="http://places2.csail.mit.edu/",
    )

  def _split_generators(self, dl_manager):
    output_archives = dl_manager.download(
        {
            "train": urllib.parse.urljoin(_BASE_URL, _TRAIN_URL),
        }
    )
    annotation_path = dl_manager.download_and_extract(
        urllib.parse.urljoin(_BASE_URL, _FILE_ANNOTATION_URL)
    )

    return {
        "train": self._generate_examples(
            archive=dl_manager.iter_archive(output_archives["train"]),
            path_prefix="Images256",
            split_name="train",
            annotation_path=os.path.join(
                annotation_path, "filelist_placesfull/imagelist_placesfull.txt"
            ),
        )
    }

  def _generate_examples(
      self, archive, path_prefix, split_name, annotation_path
  ):
    with epath.Path(annotation_path).open() as f:
      if split_name == "test":
        # Test split doesn't have labels assigned.
        file_to_class = {x[0]: -1 for x in csv.reader(f, delimiter=" ")}
      else:
        file_to_class = {x[0]: int(x[1]) for x in csv.reader(f, delimiter=" ")}

    for fname, fobj in archive:
      fname = fname.replace("\\", "/")  # For Windows compatibility.
      assert fname.startswith(path_prefix)
      # The filenames in annotations for train start with "/" while the names
      # for test and validation do not have a leading "/", so we chop
      # differently.
      chop = len(path_prefix) if split_name == "train" else len(path_prefix) + 1
      key = fname[chop:]

      class_id = file_to_class[key]
      yield fname, {"image": fobj, "label": class_id, "filename": fname}



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Convert a prepared dataset to another file format.

For example, you have a dataset in tfrecord format, but want to convert it to
ArrayRecord.

Instructions:

```
python tensorflow_datasets/scripts/convert_format.py \
  --dataset_version_dir=/data/my_dataset/config/1.2.3 \
  --out_file_format=array_record \
  --out_dir=/data/array_record/my_dataset/config/1.2.3
```

If the dataset is big, you may want to use Beam to convert it. You can do that
by adding `--use_beam`.

"""

from absl import app
from absl import flags
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.scripts.cli import convert_format_utils


_ROOT_DATA_DIR = flags.DEFINE_string(
    'root_data_dir',
    required=False,
    help=(
        'Root data dir that contains all datasets. All datasets and all their'
        ' configs and versions that are in this folder will be converted.'
    ),
    default=None,
)
_DATASET_DIR = flags.DEFINE_string(
    'dataset_dir',
    required=False,
    help=(
        'Path where the dataset to be converted is located. Converts all'
        ' configs and versions in this folder.'
    ),
    default=None,
)
_DATASET_VERSION_DIR = flags.DEFINE_list(
    'dataset_version_dir',
    required=False,
    help=(
        'Path where the dataset to be converted is located. Should include'
        ' config and version. Can also be a comma-separated list of paths. If'
        ' multiple paths are specified, `--out_dir` should not be specified,'
        ' since each dataset will be converted in the same directory as the'
        ' input dataset.'
    ),
    default=None,
)

_OUT_FILE_FORMAT = flags.DEFINE_enum_class(
    'out_file_format',
    enum_class=file_adapters.FileFormat,
    required=True,
    help='File format to convert the dataset to.',
    default=None,
)

_OUT_DIR = flags.DEFINE_string(
    'out_dir',
    required=False,
    help=(
        'Path where the converted dataset will be stored. Should include the'
        ' config and version, e.g. `/data/dataset_name/config/1.2.3`. If not'
        ' specified, the converted shards will be stored in the same directory'
        ' as the input dataset.'
    ),
    default=None,
)

_USE_BEAM = flags.DEFINE_bool(
    'use_beam',
    default=False,
    help='Whether to use beam to convert the dataset.',
)

_NUM_WORKERS = flags.DEFINE_integer(
    'num_workers',
    default=8,
    help='Number of workers to use if `use_beam` is `False`.',
)


_OVERWRITE = flags.DEFINE_bool(
    'overwrite',
    default=False,
    help='Whether to overwrite the output folder.',
)


def main(_):

  convert_format_utils.convert_dataset(
      root_data_dir=_ROOT_DATA_DIR.value,
      dataset_dir=_DATASET_DIR.value,
      dataset_version_dir=_DATASET_VERSION_DIR.value,
      out_file_format=_OUT_FILE_FORMAT.value,
      out_dir=_OUT_DIR.value,
      use_beam=_USE_BEAM.value,
      overwrite=_OVERWRITE.value,
      num_workers=_NUM_WORKERS.value,
  )


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""smartwatch_gestures dummy data generator."""
import random

from absl import app
from etils import epath
import tensorflow_datasets as tfds

NUMBER_OF_EXAMPLES = 3


def _output_dir():
  return (
      tfds.core.tfds_path()
      / 'time_series'
      / 'smartwatch_gestures'
      / 'dummy_data'
      / 'gestures-dataset'
  )


def _dummy_file(participant: int, gesture: int, attempt: int) -> epath.Path:
  """Return path to fake data file."""
  participant = f'U{participant:02}'
  gesture = f'{gesture:02}'
  attempt = f'{attempt:02}.txt'
  return _output_dir() / participant / gesture / attempt


def _init_time(num_digits: int) -> int:
  """Return rand time int."""
  return random.randint(0, (10**num_digits) - 1)


def _init_accel() -> float:
  """Return random accel float."""
  sign = 1 if random.random() < 0.5 else -1
  return sign * random.random() * 13


def _inc_time(t: int, type_str: str = 'nano') -> int:
  """Increment time at aprrox rate of 10Hz."""
  return int(t + 1e8) if type_str == 'nano' else t + 100


def _inc_accel(x: float) -> float:
  sign = 1 if random.random() < 0.5 else -1
  return x + sign * random.random() * 2


def _generate_data():
  """Generate dummy data."""
  for i in range(NUMBER_OF_EXAMPLES):
    fpath = _dummy_file(99, 1, i)
    if not fpath.exists():
      fpath.parent.mkdir(parents=True, exist_ok=True)
    else:
      fpath.unlink()

    t_milli = _init_time(13)
    t_nanos = _init_time(14)
    t_event = _init_time(13)
    accel_x = _init_accel()
    accel_y = _init_accel()
    accel_z = _init_accel()

    with fpath.open('w') as f:
      f.write(
          f'{t_milli:013} {t_nanos:014} {t_event:013}'
          f' {accel_x:.6f} {accel_y:.6f} {accel_z:.6f}\n'
      )

      for _ in range(random.randint(11, 51)):
        t_milli = _inc_time(t_milli, 'milli')
        t_nanos = _inc_time(t_nanos)
        t_event = _inc_time(t_event)
        accel_x = _inc_accel(accel_x)
        accel_y = _inc_accel(accel_y)
        accel_z = _inc_accel(accel_z)

        f.write(
            f'{t_milli:013} {t_nanos:014} {t_event:013}'
            f' {accel_x:.6f} {accel_y:.6f} {accel_z:.6f}\n'
        )


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  _generate_data()


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""smart_buildings dataset."""

from collections.abc import Iterable
import datetime
from typing import Any

from absl import logging
from etils import epath
from google.protobuf import json_format
import numpy as np
import pandas as pd
from tensorflow_datasets.core.utils.lazy_imports_utils import apache_beam as beam
from tensorflow_datasets.datasets.smart_buildings import controller_reader
import tensorflow_datasets.public_api as tfds

# The years in the dataset.
YEARS = [19, 20, 21, 22, 23, 24]

_MAX_EXAMPLES_PER_DAY = 24 * 12  # 288 per day

_REWARD_RESPONSES = [
    'agentRewardValue',
    'productivityReward',
    'electricityEnergyCost',
    'carbonEmitted',
    'productivityWeight',
    'energyCostWeight',
    'carbonEmissionWeight',
    'personProductivity',
    'totalOccupancy',
    'rewardScale',
    'productivityRegret',
    'normalizedProductivityRegret',
    'normalizedEnergyCost',
    'normalizedCarbonEmission',
    'naturalGasEnergyCost',
]


def to_ns_timestamp(proto_timestamp) -> int:
  """Converts a protobuf.Timestamp to number of nanoseconds since epoch."""
  return proto_timestamp.seconds * 1e9 + proto_timestamp.nanos


class BuilderConfig(tfds.core.BuilderConfig):
  building: str


def _make_building_config(building: str) -> BuilderConfig:
  config = BuilderConfig(
      name=building,
      version=tfds.core.utils.version.Version('1.0.0'),
      description=f'Building {building}',
  )
  config.building = building
  return config


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for smart_buildings dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      _make_building_config(building='sb1'),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'observation': tfds.features.FeaturesDict({
                'timestamp': tfds.features.Text(),
                'request': tfds.features.FeaturesDict({
                    'timestamp': tfds.features.Text(),
                    'singleObservationRequests': tfds.features.Sequence(
                        tfds.features.FeaturesDict({
                            'deviceId': tfds.features.Text(),
                            'measurementName': tfds.features.Text(),
                        })
                    ),
                }),
                'singleObservationResponses': tfds.features.Sequence(
                    tfds.features.FeaturesDict({
                        'singleObservationRequest': tfds.features.FeaturesDict({
                            'deviceId': tfds.features.Text(),
                            'measurementName': tfds.features.Text(),
                        }),
                        'timestamp': tfds.features.Text(),
                        'continuousValue': tfds.features.Scalar(
                            dtype=np.float32
                        ),
                        'observationValid': tfds.features.Text(),
                    })
                ),
            }),
            'action': tfds.features.FeaturesDict({
                'timestamp': tfds.features.Text(),
                'request': tfds.features.FeaturesDict({
                    'timestamp': tfds.features.Text(),
                    'singleActionRequests': tfds.features.Sequence(
                        tfds.features.FeaturesDict({
                            'deviceId': tfds.features.Text(),
                            'setpointName': tfds.features.Text(),
                            'continuousValue': tfds.features.Scalar(
                                dtype=np.float32
                            ),
                        })
                    ),
                }),
                'singleActionResponses': tfds.features.Sequence(
                    tfds.features.FeaturesDict({
                        'request': tfds.features.FeaturesDict({
                            'deviceId': tfds.features.Text(),
                            'setpointName': tfds.features.Text(),
                            'continuousValue': tfds.features.Scalar(
                                dtype=np.float32
                            ),
                        }),
                        'responseType': tfds.features.Text(),
                        'additionalInfo': tfds.features.Text(),
                    }),
                ),
            }),
            'reward': tfds.features.FeaturesDict({
                'productivityWeight': tfds.features.Scalar(dtype=np.float32),
                'energyCostWeight': tfds.features.Scalar(dtype=np.float32),
                'carbonEmissionWeight': tfds.features.Scalar(dtype=np.float32),
                'rewardScale': tfds.features.Scalar(dtype=np.float32),
                'startTimestamp': tfds.features.Text(),
                'endTimestamp': tfds.features.Text(),
                'agentRewardValue': tfds.features.Scalar(dtype=np.float32),
                'carbonEmitted': tfds.features.Scalar(dtype=np.float32),
                'electricityEnergyCost': tfds.features.Scalar(dtype=np.float32),
                'normalizedCarbonEmission': tfds.features.Scalar(
                    dtype=np.float32
                ),
                'normalizedEnergyCost': tfds.features.Scalar(dtype=np.float32),
                'normalizedProductivityRegret': tfds.features.Scalar(
                    dtype=np.float32
                ),
                'personProductivity': tfds.features.Scalar(dtype=np.float32),
                'productivityRegret': tfds.features.Scalar(dtype=np.float32),
                'productivityReward': tfds.features.Scalar(dtype=np.float32),
                'totalOccupancy': tfds.features.Scalar(dtype=np.float32),
                'naturalGasEnergyCost': tfds.features.Scalar(dtype=np.float32),
            }),
        }),
        homepage='https://github.com/google/sbsim',
        disable_shuffling=True,  # our dataset needs to be in order
    )

  def _split_generators(
      self, dl_manager: tfds.download.DownloadManager, pipeline
  ):
    """Download the data and define splits."""
    building = self.builder_config.building
    building_upper = building.upper()

    path_by_year: dict[int, epath.Path] = dl_manager.download_and_extract({
        year: f'https://storage.googleapis.com/gresearch/smart_buildings_dataset/{building_upper}/{building_upper}_{year}.zip'
        for year in YEARS
    })

    splits_dict = {}
    for year, path in path_by_year.items():
      splits_dict[f'{building}_{year}'] = self._generate_examples(
          path=path / 'dataset' / building_upper / str(year),
          year=year,
          pipeline=pipeline,
      )
    return splits_dict

  def _generate_examples(self, path: epath.Path, year: int, pipeline):
    """Yields examples."""
    logging.info('Processing year %d in path %s', year, path)

    start_date = datetime.date(2000 + year, 1, 1)
    end_date = datetime.date(2000 + year + 1, 1, 1)
    all_dates = pd.date_range(start_date, end_date)

    return (
        pipeline
        | f'CreateDates_{year}' >> beam.Create(enumerate(all_dates))
        | f'ProcessDate_{year}' >> beam.FlatMap(process_date, path=path)
        | f'Reshuffle_{year}' >> beam.Reshuffle()
    )


def process_date(
    day_index_and_start_time: tuple[int, pd.Timestamp],
    path: epath.Path,
) -> Iterable[tuple[int, dict[str, Any]]]:
  """Process a single date."""
  day_index, start_time = day_index_and_start_time
  end_time = start_time + pd.Timedelta(hours=23)
  key_offset = day_index * _MAX_EXAMPLES_PER_DAY

  reader = controller_reader.ProtoReader(path)
  observation_responses = reader.read_observation_responses(
      start_time, end_time
  )
  action_responses = reader.read_action_responses(start_time, end_time)
  reward_responses = reader.read_reward_responses(start_time, end_time)

  # make sure all are sorted by data
  observation_responses = sorted(
      observation_responses,
      key=lambda o: to_ns_timestamp(o.timestamp),
  )
  action_responses = sorted(
      action_responses,
      key=lambda o: to_ns_timestamp(o.timestamp),
  )
  reward_responses = sorted(
      reward_responses,
      key=lambda o: to_ns_timestamp(o.start_timestamp),
  )

  if len(observation_responses) > _MAX_EXAMPLES_PER_DAY:
    raise ValueError(
        f'Too many observation responses for date {start_time}: '
        f'{len(observation_responses)} > {_MAX_EXAMPLES_PER_DAY}'
    )

  for i in range(len(observation_responses)):
    observation_response = json_format.MessageToDict(observation_responses[i])
    action_response = json_format.MessageToDict(action_responses[i])
    reward_response = json_format.MessageToDict(reward_responses[i])

    # fill missing values
    for single_observation_response in observation_response[
        'singleObservationResponses'
    ]:
      if 'continuousValue' not in single_observation_response:
        single_observation_response['continuousValue'] = 0.0
      if 'observationValid' not in single_observation_response:
        single_observation_response['observationValid'] = 'False'
      else:
        single_observation_response['observationValid'] = str(
            single_observation_response['observationValid']
        )  # parse bool to string
      if 'timestamp' not in single_observation_response:
        single_observation_response['timestamp'] = observation_response[
            'timestamp'
        ]
    for val in _REWARD_RESPONSES:
      if val not in reward_response:
        reward_response[val] = -1  # sentinal value

    beam.metrics.Metrics.counter(f'date_{start_time}', 'example_count').inc()
    key = key_offset + i
    yield key, {
        'observation': observation_response,
        'action': action_response,
        'reward': reward_response,
    }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for ble_wind_field."""

from unittest import mock

from tensorflow_datasets.datasets.ble_wind_field import ble_wind_field_dataset_builder
import tensorflow_datasets.public_api as tfds


class BLEWindFieldTest(tfds.testing.DatasetBuilderTestCase):
  DATASET_CLASS = ble_wind_field_dataset_builder.Builder
  SPLITS = {'train': 4}

  def setUp(self):
    super().setUp()
    # Patch BLEWindField's GCS_URL and GCS_FILENAME class attributes to
    # point to the placeholder Zarr array instead. Also patch
    # tfds.core.lazy_imports.gcsfs_store to turn it into an identity function,
    # because the original function doesn't work with local paths (i.e. not of
    # the form 'gs://*').
    module_name = self.DATASET_CLASS.__module__
    class_name = self.DATASET_CLASS.__name__
    patchers = [
        mock.patch(f'{module_name}.{class_name}.GCS_URL', self.dummy_data),
        mock.patch(f'{module_name}.{class_name}.GCS_FILENAME', 'array.zarr'),
        mock.patch(
            'tensorflow_datasets.public_api.core.lazy_imports.gcsfs_store',
            lambda s: s,
        ),
    ]
    for patcher in patchers:
      patcher.start()
    self.patchers.extend(patchers)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""All RT-X datasets."""

from tensorflow_datasets.robotics.rtx.rtx import AlohaMobile
from tensorflow_datasets.robotics.rtx.rtx import AsuTableTopConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import AustinBudsDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import AustinSailorDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import AustinSiriusDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import BcZ
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyAutolabUr5
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyCableRouting
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyFanucManipulation
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyGnmCoryHall
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyGnmRecon
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyGnmSacSon
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyMvpConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import BerkeleyRptConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import Bridge
from tensorflow_datasets.robotics.rtx.rtx import BridgeDataMsr
from tensorflow_datasets.robotics.rtx.rtx import CmuFrankaExplorationDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import CmuPlayFusion
from tensorflow_datasets.robotics.rtx.rtx import CmuStretch
from tensorflow_datasets.robotics.rtx.rtx import ColumbiaCairlabPushtReal
from tensorflow_datasets.robotics.rtx.rtx import ConqHoseManipulation
from tensorflow_datasets.robotics.rtx.rtx import DlrEdanSharedControlConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import DlrSaraGridClampConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import DlrSaraPourConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import Dobbe
from tensorflow_datasets.robotics.rtx.rtx import EthAgentAffordances
from tensorflow_datasets.robotics.rtx.rtx import Fmb
from tensorflow_datasets.robotics.rtx.rtx import Fractal20220817Data
from tensorflow_datasets.robotics.rtx.rtx import IamlabCmuPickupInsertConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import ImperialcollegeSawyerWristCam
from tensorflow_datasets.robotics.rtx.rtx import IoAiTech
from tensorflow_datasets.robotics.rtx.rtx import JacoPlay
from tensorflow_datasets.robotics.rtx.rtx import KaistNonprehensileConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import Kuka
from tensorflow_datasets.robotics.rtx.rtx import ManiskillDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import MimicPlay
from tensorflow_datasets.robotics.rtx.rtx import NyuDoorOpeningSurprisingEffectiveness
from tensorflow_datasets.robotics.rtx.rtx import NyuFrankaPlayDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import NyuRotDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import PlexRobosuite
from tensorflow_datasets.robotics.rtx.rtx import RoboSet
from tensorflow_datasets.robotics.rtx.rtx import Roboturk
from tensorflow_datasets.robotics.rtx.rtx import SpocRobot
from tensorflow_datasets.robotics.rtx.rtx import StanfordHydraDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import StanfordKukaMultimodalDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import StanfordMaskVitConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import StanfordRobocookConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import TacoPlay
from tensorflow_datasets.robotics.rtx.rtx import Tidybot
from tensorflow_datasets.robotics.rtx.rtx import TokyoULsmoConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import Toto
from tensorflow_datasets.robotics.rtx.rtx import UcsdKitchenDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UcsdPickAndPlaceDatasetConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UiucD3field
from tensorflow_datasets.robotics.rtx.rtx import UscClothSimConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UtaustinMutex
from tensorflow_datasets.robotics.rtx.rtx import UtokyoPr2OpeningFridgeConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UtokyoPr2TabletopManipulationConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UtokyoSaytapConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UtokyoXarmBimanualConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import UtokyoXarmPickAndPlaceConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import VimaConvertedExternallyToRlds
from tensorflow_datasets.robotics.rtx.rtx import Viola



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RefCoco datasets."""

from __future__ import annotations

import collections
import io
import json
import operator

import numpy as np
import tensorflow_datasets.public_api as tfds


def _build_bbox(image_info, x, y, width, height):
  """Calculates the coordinates of a bbox."""
  return tfds.features.BBox(
      ymin=y / image_info['height'],
      xmin=x / image_info['width'],
      ymax=(y + height) / image_info['height'],
      xmax=(x + width) / image_info['width'],
  )


def _compute_mask(image_info, ann):
  h, w = image_info['height'], image_info['width']
  rles = tfds.core.lazy_imports.pycocotools.frPyObjects(
      ann['segmentation'], h, w
  )
  rle = tfds.core.lazy_imports.pycocotools.merge(rles)
  return tfds.core.lazy_imports.pycocotools.decode(rle)


def _extract_annotation(ann, image_info):
  """Extracts the bounding box annotation information."""
  return {
      'id': ann['id'],
      'area': ann['area'],
      'bbox': _build_bbox(image_info, *ann['bbox']),
      'label': ann['category_id'],
  }


def _generate_examples(refcoco_json, dataset, dataset_partition, split):
  """Generates examples of images and its refexps & ground truth bboxes.

  Args:
    refcoco_json: contents of the annotation file.
    dataset: str specifying the dataset (refcoco, refcoco+, refcocog)
    dataset_partition: str specifying the partition for the dataset
    split: str specifying the split of the dataset_partition

  Yields:
    image_id and example tuple
  """
  refcoco_anns = refcoco_json['ref']
  coco_anns = refcoco_json['coco_anns']

  # Collect all referring expressions for a given image.
  imageid2annref = collections.defaultdict(list)
  for r in refcoco_anns:
    if (
        r['dataset'] == dataset
        and r['dataset_partition'] == dataset_partition
        and r['split'] == split
    ):
      imageid2annref[r['image_id']].append(r)

  # Process all the referring expressions and ground truth annotations for
  # a given COCO image.
  for image_id in sorted(imageid2annref.keys()):
    coco_image = coco_anns[str(image_id)]
    image_info = coco_image['info']
    example = {
        'image_filename': image_info['file_name'],
        'image/id': image_id,
        'coco_annotations': [],
        'objects': [],
    }

    # Collect ground truth bboxes.
    for ann in sorted(coco_image['anns'], key=operator.itemgetter('id')):
      example['coco_annotations'].append(_extract_annotation(ann, image_info))

    # Collect referring expressions.
    for r in sorted(
        imageid2annref[image_id], key=operator.itemgetter('ref_id')
    ):
      obj = _extract_annotation(r['ann'], image_info)

      refexp = []
      for s in sorted(r['sentences'], key=operator.itemgetter('sent_id')):
        refexp.append({
            'raw': s['raw'],
            'refexp_id': s['sent_id'],
        })

      # Match the referring expression to its corresponding bbox in the ground
      # truth list.
      gt_box_index = [
          i
          for i, v in enumerate(example['coco_annotations'])
          if v['id'] == r['ann']['id']
      ]
      if len(gt_box_index) != 1:
        raise ValueError(f'gt_box_index does not have length 1: {gt_box_index}')
      gt_box_index = gt_box_index[0]

      mask = _compute_mask(image_info, r['ann']).astype(np.bool_)
      mask_buf = io.BytesIO()
      tfds.core.lazy_imports.PIL_Image.fromarray(mask).save(
          mask_buf, format='png'
      )
      mask_buf.seek(0)

      obj.update({
          'refexp': refexp,
          'gt_box_index': gt_box_index,
          'mask': mask_buf,
      })
      example['objects'].append(obj)

    yield image_id, example


class RefCocoConfig(tfds.core.BuilderConfig):
  """Config to specify each RefCoco variant."""

  def __init__(self, dataset, dataset_partition, **kwargs):
    name = f'{dataset}_{dataset_partition}'
    super(RefCocoConfig, self).__init__(name=name, **kwargs)
    self.dataset = dataset
    self.dataset_partition = dataset_partition


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for RefCoco datasets."""

  VERSION = tfds.core.Version('1.1.0')
  RELEASE_NOTES = {
      '1.1.0': 'Added masks.',
      '1.0.0': 'Initial release.',
  }

  MANUAL_DOWNLOAD_INSTRUCTIONS = """
  1. Follow the instructions in https://github.com/lichengunc/refer and
  download the annotations and the images, matching the data/ directory
  specified in the repo.

  2. Follow the instructions of PythonAPI in
  https://github.com/cocodataset/cocoapi to get pycocotools and the
  instances_train2014 annotations file from https://cocodataset.org/#download

  3. Add both refer.py from (1) and pycocotools from (2) to your PYTHONPATH.

  4. Run manual_download_process.py to generate refcoco.json, replacing
  `ref_data_root`, `coco_annotations_file`, and `out_file` with the values
  corresponding to where you have downloaded / want to save these files.
  Note that manual_download_process.py can be found in the TFDS repository.

  5. Download the COCO training set from https://cocodataset.org/#download
  and stick it into a folder called `coco_train2014/`. Move `refcoco.json`
  to the same level as `coco_train2014`.

  6. Follow the standard manual download instructions.
  """

  BUILDER_CONFIGS = [
      RefCocoConfig(dataset='refcoco', dataset_partition='unc'),
      RefCocoConfig(dataset='refcoco', dataset_partition='google'),
      RefCocoConfig(dataset='refcocoplus', dataset_partition='unc'),
      RefCocoConfig(dataset='refcocog', dataset_partition='google'),
      RefCocoConfig(dataset='refcocog', dataset_partition='umd'),
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        homepage='https://github.com/lichengunc/refer',
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'image/id': np.int64,
            'objects': tfds.features.Sequence({
                'id': np.int64,
                'area': np.int64,
                'bbox': tfds.features.BBoxFeature(),
                'label': np.int64,
                'gt_box_index': np.int64,
                'mask': tfds.features.Image(encoding_format='png'),
                'refexp': tfds.features.Sequence({
                    'refexp_id': np.int64,
                    'raw': tfds.features.Text(),
                }),
            }),
            'coco_annotations': tfds.features.Sequence({
                'id': np.int64,
                'area': np.int64,
                'bbox': tfds.features.BBoxFeature(),
                'label': np.int64,
            }),
        }),
        supervised_keys=None,
    )

  def _split_generators(self, dl_manager):
    allowed_splits = {
        ('refcoco', 'google'): [
            tfds.Split.TRAIN,
            tfds.Split.VALIDATION,
            tfds.Split.TEST,
        ],
        ('refcoco', 'unc'): [
            tfds.Split.TRAIN,
            tfds.Split.VALIDATION,
            'testA',
            'testB',
        ],
        ('refcocoplus', 'unc'): [
            tfds.Split.TRAIN,
            tfds.Split.VALIDATION,
            'testA',
            'testB',
        ],
        ('refcocog', 'google'): [tfds.Split.TRAIN, tfds.Split.VALIDATION],
        ('refcocog', 'umd'): [
            tfds.Split.TRAIN,
            tfds.Split.VALIDATION,
            tfds.Split.TEST,
        ],
    }
    bc = self.builder_config
    splits = allowed_splits[(bc.dataset, bc.dataset_partition)]

    return {
        split: self._generate_examples(
            bc.dataset, bc.dataset_partition, split, dl_manager
        )
        for split in splits
    }

  def _generate_examples(self, dataset, dataset_partition, split, dl_manager):
    refcoco_json = json.loads(
        (dl_manager.manual_dir / 'refcoco.json').read_text()
    )
    coco_dir = dl_manager.manual_dir / 'coco_train2014'

    if dataset == 'refcocoplus':
      dataset = 'refcoco+'
    if split == tfds.Split.VALIDATION:
      split = 'val'

    for image_id, example in _generate_examples(
        refcoco_json, dataset, dataset_partition, split
    ):
      example['image'] = coco_dir / example['image_filename']
      del example['image_filename']
      yield image_id, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""scitail dataset."""

import csv

from etils import epath
import tensorflow_datasets.public_api as tfds

_URL = 'http://data.allenai.org.s3.amazonaws.com/downloads/SciTailV1.1.zip'
_SCITAIL_DIR = 'SciTailV1.1'
_TSV_DIR = 'tsv_format'


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for scitail dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    # Specifies the tfds.core.DatasetInfo object
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'premise': tfds.features.Text(),
            'hypothesis': tfds.features.Text(),
            # Label indicates whether the premise entails or implies the
            # hypothesis.
            'label': tfds.features.ClassLabel(names=['entails', 'neutral']),
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=None,  # Set to `None` to disable
        homepage='https://allenai.org/data/scitail',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    # Downloads the data and defines the splits
    data_dir = dl_manager.download_and_extract(_URL)

    # Returns the Dict[split names, Iterator[Key, Example]]
    return {
        tfds.Split.TRAIN: self._generate_examples(
            path=data_dir / _SCITAIL_DIR / _TSV_DIR / 'scitail_1.0_train.tsv'
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            path=data_dir / _SCITAIL_DIR / _TSV_DIR / 'scitail_1.0_dev.tsv'
        ),
        tfds.Split.TEST: self._generate_examples(
            path=data_dir / _SCITAIL_DIR / _TSV_DIR / 'scitail_1.0_test.tsv'
        ),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    with epath.Path(path).open() as f:
      data = csv.reader(f, delimiter='\t')
      for id_, row in enumerate(data):
        yield id_, {'premise': row[0], 'hypothesis': row[1], 'label': row[2]}



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for assin2_utils."""
from tensorflow_datasets import testing
from tensorflow_datasets.datasets.assin2.assin2_utils import Pair
from tensorflow_datasets.datasets.assin2.assin2_utils import parse_xml_string

# pylint: disable=line-too-long
_CAND_XML_STR = """\
<?xml version='1.0' encoding='utf-8'?>
<entailment-corpus>
         <pair entailment="Entailment" id="1" similarity="4.5">
             <t>Uma criança risonha está segurando uma pistola de água e sendo espirrada com água</t>
             <h>Uma criança está segurando uma pistola de água</h>
         </pair>
         <pair entailment="Entailment" id="2" similarity="4.5">
             <t>Os homens estão cuidadosamente colocando as malas no porta-malas de um carro</t>
             <h>Os homens estão colocando bagagens dentro do porta-malas de um carro</h>
         </pair>
         <pair entailment="Entailment" id="3" similarity="4.7">
             <t>Uma pessoa tem cabelo loiro e esvoaçante e está tocando violão</t>
             <h>Um guitarrista tem cabelo loiro e esvoaçante</h>
         </pair>
         <pair entailment="Entailment" id="4" similarity="4.7">
             <t>Batatas estão sendo fatiadas por um homem</t>
             <h>O homem está fatiando a batata</h>
         </pair>
         <pair entailment="Entailment" id="5" similarity="4.9">
             <t>Um caminhão está descendo rapidamente um morro</t>
             <h>Um caminhão está rapidamente descendo o morro</h>
         </pair>
</entailment-corpus>
"""

# Same data from ./dummy_data/assin2-train-only.xml
_EXPECTED_PAIRS = [
    Pair(
        text=(
            'Uma criança risonha está segurando uma pistola de água e sendo'
            ' espirrada com água'
        ),
        hypothesis='Uma criança está segurando uma pistola de água',
        id=1,
        entailment='Entailment',
        similarity=4.5,
    ),
    Pair(
        text=(
            'Os homens estão cuidadosamente colocando as malas no porta-malas'
            ' de um carro'
        ),
        hypothesis=(
            'Os homens estão colocando bagagens dentro do porta-malas de um'
            ' carro'
        ),
        id=2,
        entailment='Entailment',
        similarity=4.5,
    ),
    Pair(
        text='Uma pessoa tem cabelo loiro e esvoaçante e está tocando violão',
        hypothesis='Um guitarrista tem cabelo loiro e esvoaçante',
        id=3,
        entailment='Entailment',
        similarity=4.7,
    ),
    Pair(
        text='Batatas estão sendo fatiadas por um homem',
        hypothesis='O homem está fatiando a batata',
        id=4,
        entailment='Entailment',
        similarity=4.7,
    ),
    Pair(
        text='Um caminhão está descendo rapidamente um morro',
        hypothesis='Um caminhão está rapidamente descendo o morro',
        id=5,
        entailment='Entailment',
        similarity=4.9,
    ),
]


class Assin2UtilsTest(testing.TestCase):

  def test_parse_xml_string(self):
    cand_pairs = parse_xml_string(_CAND_XML_STR)
    self.assertLen(cand_pairs, len(_EXPECTED_PAIRS))
    self.assertEqual(_EXPECTED_PAIRS, cand_pairs)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Historical wind field dataset for the Balloon Learning Environment."""

import concurrent.futures
import dataclasses
import itertools
from typing import Optional

import numpy as np
import tensorflow_datasets.public_api as tfds


@dataclasses.dataclass
class BLEWindFieldConfig(tfds.core.BuilderConfig):
  num_fields: Optional[int] = None


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for the ble_wind_field dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      BLEWindFieldConfig(
          name='full',
          description='The entire historical wind field dataset.',
          num_fields=None,
      ),
      BLEWindFieldConfig(
          name='small',
          description='Small sample of 256 fields from the dataset.',
          num_fields=256,
      ),
  ]
  GCS_URL = tfds.core.Path('gs://ble-public/downloads')
  GCS_FILENAME = 'historical_wind_fields.zarr'

  def _info(self) -> tfds.core.DatasetInfo:
    """Dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                'field': tfds.features.Tensor(
                    shape=(21, 21, 10, 9, 2),
                    dtype=np.float32,
                    encoding=tfds.features.Encoding.ZLIB,
                ),
            }
        ),
        supervised_keys=None,
        homepage='https://github.com/google/balloon-learning-environment',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    return {'train': self._generate_examples()}

  def _generate_examples(self):
    zarr = tfds.core.lazy_imports.zarr
    gcsfs_store = tfds.core.lazy_imports.gcsfs_store
    zarr_array = zarr.open_array(
        store=gcsfs_store(f'{self.GCS_URL}/{self.GCS_FILENAME}'),
        mode='r',
    )

    # During normal execution we don't expect `self.builder_config.num_fields`
    # to have a value larger than `zarr_array.shape[0]`, but for unit tests the
    # Zarr file used has a very small number of fields, so we take the minimum
    # to avoid trying to load more examples than available.
    num_fields = min(
        self.builder_config.num_fields or zarr_array.shape[0],
        zarr_array.shape[0],
    )

    # Zarr arrays are stored as compressed chunks on disk, and by default
    # read/write operations require to load and decompress entire chunks, even
    # if a single element in the chunk is accessed. The most efficient way of
    # iterating over elements is to load entire chunks and iterate over the
    # chunks' elements. The data can be chunked across all axes, but for
    # simplicity we assume that it's only chunked across the batch axis (which
    # is how the historical wind field array is organized).
    chunk_length = zarr_array.chunks[0]
    num_full_chunks = num_fields // chunk_length
    remainder = num_fields % chunk_length

    slices = (
        slice(i * chunk_length, (i + 1) * chunk_length)
        for i in range(num_full_chunks)
    )
    if remainder:
      start = num_full_chunks * chunk_length
      slices = itertools.chain(slices, (slice(start, start + remainder),))

    keys = itertools.count()
    # We pipeline loading the data and writing it to tfrecords files using a
    # thread that loads chunks asynchronously.
    with concurrent.futures.ThreadPoolExecutor() as executor:
      load_fn = lambda next_slice: zarr_array[next_slice]
      # Keep at most 10 submitted loading tasks to reduce resource consumption.
      futures = {
          executor.submit(load_fn, next_slice)
          for next_slice in itertools.islice(slices, 10)
      }
      while futures:
        done, futures = concurrent.futures.wait(
            futures, return_when=concurrent.futures.FIRST_COMPLETED
        )
        # Top up the submitted loading tasks.
        for next_slice in itertools.islice(slices, len(done)):
          futures.add(executor.submit(load_fn, next_slice))

        # Yield the loaded slices.
        for future in done:
          for field in future.result():
            yield next(keys), {'field': field}



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for file_adapters."""

import os
import pathlib
from typing import Type, TypeAlias

from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming


FileFormat: TypeAlias = file_adapters.FileFormat


def test_batched():
  assert list(file_adapters._batched(range(10), 5)) == [
      [0, 1, 2, 3, 4],
      [5, 6, 7, 8, 9],
  ]
  assert list(file_adapters._batched(range(10), 100)) == [
      [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
  ]
  assert not list(file_adapters._batched(range(10), 0))


def test_is_example_file():
  assert file_adapters.is_example_file('example1.tfrecord')
  assert file_adapters.is_example_file('example1.riegeli')
  assert file_adapters.is_example_file('example1.tfrecord-00000-of-00001')
  assert not file_adapters.is_example_file('example1.info')


@pytest.mark.parametrize(
    'format_suffix,file_format',
    (
        ('array_record', FileFormat.ARRAY_RECORD),
        ('parquet', FileFormat.PARQUET),
        ('riegeli', FileFormat.RIEGELI),
        ('tfrecord', FileFormat.TFRECORD),
    ),
)
def test_format_suffix(format_suffix, file_format):
  assert (
      file_adapters.ADAPTER_FOR_FORMAT[file_format].FILE_SUFFIX == format_suffix
  )


@pytest.mark.parametrize(
    'file_format',
    [
        file_format
        for file_format in FileFormat
        if file_format in FileFormat.with_tf_data()
        and file_format != FileFormat.RIEGELI
    ],
)
@pytest.mark.parametrize(
    'builder_cls',
    [
        testing.DummyDataset,
        testing.DummyBeamDataset,
    ],
)
def test_read_write(
    tmp_path: pathlib.Path,
    builder_cls: Type[dataset_builder.DatasetBuilder],
    file_format: FileFormat,
):
  builder = builder_cls(data_dir=tmp_path, file_format=file_format)
  builder.download_and_prepare()
  ds = builder.as_dataset(split='train')
  assert list(ds.as_numpy_iterator()) == [{'id': i} for i in range(3)]


def test_as_dataset_fails_for_array_record(
    tmp_path: pathlib.Path,
):
  builder = testing.DummyDataset(
      data_dir=tmp_path, file_format=FileFormat.ARRAY_RECORD
  )
  builder.download_and_prepare()
  with pytest.raises(
      NotImplementedError, match='not implemented for ArrayRecord files'
  ):
    ds = builder.as_dataset(split='train')


@pytest.mark.parametrize(
    'format_enum_value,file_format',
    (
        ('array_record', FileFormat.ARRAY_RECORD),
        ('parquet', FileFormat.PARQUET),
        ('riegeli', FileFormat.RIEGELI),
        ('tfrecord', FileFormat.TFRECORD),
    ),
)
def test_prase_file_format(format_enum_value, file_format):
  assert FileFormat.from_value(format_enum_value) == file_format
  assert FileFormat.from_value(file_format) == file_format


@pytest.mark.parametrize(
    'path,file_format,expected_path',
    (
        (
            '/tmp/dataset-train.tfrecord-00000-of-00001',
            file_adapters.FileFormat.TFRECORD,
            '/tmp/dataset-train.tfrecord-00000-of-00001',
        ),
        (
            '/tmp/dataset-train.riegeli-00000-of-00001',
            file_adapters.FileFormat.TFRECORD,
            '/tmp/dataset-train.tfrecord-00000-of-00001',
        ),
        (
            '/tmp/dataset-train.tfrecord-00000-of-00001',
            file_adapters.FileFormat.RIEGELI,
            '/tmp/dataset-train.riegeli-00000-of-00001',
        ),
    ),
)
def test_convert_path_to_file_format(path, file_format, expected_path):
  converted_path = file_adapters.convert_path_to_file_format(path, file_format)
  assert os.fspath(converted_path) == expected_path


@pytest.mark.parametrize(
    'adapter_cls',
    (
        (file_adapters.TfRecordFileAdapter),
        (file_adapters.ArrayRecordFileAdapter),
    ),
)
def test_shard_lengths(
    tmp_path: pathlib.Path, adapter_cls: file_adapters.FileAdapter
):
  file_template = naming.ShardedFileTemplate(
      data_dir=tmp_path,
      dataset_name='data',
      filetype_suffix=adapter_cls.FILE_SUFFIX,
      split='train',
  )
  tmp_path_1 = file_template.sharded_filepath(shard_index=0, num_shards=2)
  tmp_path_2 = file_template.sharded_filepath(shard_index=1, num_shards=2)
  adapter_cls.write_examples(
      tmp_path_1, [(0, b'0'), (1, b'1'), (2, b'2222'), (3, b'33333')]
  )
  adapter_cls.write_examples(tmp_path_2, [(3, b'3'), (4, b'4'), (5, b'555')])
  size_1 = epath.Path(tmp_path_1).stat().length
  size_2 = epath.Path(tmp_path_2).stat().length
  expected_shard_lengths = [(4, size_1), (3, size_2)]

  # First test without passing the number of shards explicitly.
  actual_no_num_shards = adapter_cls.shard_lengths_and_sizes(file_template)
  assert actual_no_num_shards == expected_shard_lengths, 'no num_shards passed'

  # Now test with passing the number of shards explicitly.
  actual_with_num_shards = adapter_cls.shard_lengths_and_sizes(
      file_template,
      num_shards=2,
  )
  assert actual_with_num_shards == expected_shard_lengths, 'num_shards passed'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Corrupted ImageNet2012 dataset.

Apply common corruptions to the images in ImageNet2012 dataset.
"""
from absl import logging
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet2012_dataset_builder
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
from tensorflow_datasets.image_classification import corruptions
import tensorflow_datasets.public_api as tfds

_FROST_FILEBASE = 'https://raw.githubusercontent.com/hendrycks/robustness/master/ImageNet-C/imagenet_c/imagenet_c/frost'
_FROST_FILENAMES = [f'{_FROST_FILEBASE}/frost{i}.png' for i in range(1, 4)] + [
    f'{_FROST_FILEBASE}/frost{i}.jpg' for i in range(4, 7)
]

BENCHMARK_CORRUPTIONS = [
    'gaussian_noise',
    'shot_noise',
    'impulse_noise',
    'defocus_blur',
    'glass_blur',
    'motion_blur',
    'zoom_blur',
    'snow',
    'frost',
    'fog',
    'brightness',
    'contrast',
    'elastic_transform',
    'pixelate',
    'jpeg_compression',
]

EXTRA_CORRUPTIONS = ['gaussian_blur', 'saturate', 'spatter', 'speckle_noise']

_IMAGE_SIZE = 224
_CROP_PADDING = 32


class Imagenet2012CorruptedConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Imagenet2012Corrupted."""

  def __init__(self, *, corruption_type=None, severity=1, **kwargs):
    """BuilderConfig for Imagenet2012Corrupted.

    Args:
      corruption_type: string, must be one of the items in BENCHMARK_CORRUPTIONS
        + EXTRA_CORRUPTIONS.
      severity: integer, bewteen 1 and 5.
      **kwargs: keyword arguments forwarded to super.
    """
    super(Imagenet2012CorruptedConfig, self).__init__(**kwargs)
    self.corruption_type = corruption_type
    self.severity = severity


_VERSION = tfds.core.Version('3.1.0')
_RELEASE_NOTES = {
    '0.0.1': 'Initial dataset',
    '3.1.0': (
        'Implement missing corruptions. Fix crop/resize ordering, file encoding'
    ),
}


def _make_builder_configs():
  """Construct a list of BuilderConfigs.

  Construct a list of 95 Imagenet2012CorruptedConfig objects, corresponding to
  the 15 + 4 corruption types, with each type having 5 severities.

  Returns:
    A list of 95 Imagenet2012CorruptedConfig objects.
  """
  config_list = []
  for each_corruption in BENCHMARK_CORRUPTIONS + EXTRA_CORRUPTIONS:
    for each_severity in range(1, 6):
      name_str = each_corruption + '_' + str(each_severity)
      description_str = 'corruption type = ' + each_corruption + ', severity = '
      description_str += str(each_severity)
      config_list.append(
          Imagenet2012CorruptedConfig(
              name=name_str,
              version=_VERSION,
              release_notes=_RELEASE_NOTES,
              description=description_str,
              corruption_type=each_corruption,
              severity=each_severity,
          )
      )
  return config_list


def _decode_and_center_crop(image_bytes):
  """Crops to center of image with padding then scales image size."""
  shape = tf.image.extract_jpeg_shape(image_bytes)
  image_height = shape[0]
  image_width = shape[1]

  padded_center_crop_size = tf.cast(
      (
          (_IMAGE_SIZE / (_IMAGE_SIZE + _CROP_PADDING))
          * tf.cast(tf.minimum(image_height, image_width), tf.float32)
      ),
      tf.int32,
  )

  offset_height = ((image_height - padded_center_crop_size) + 1) // 2
  offset_width = ((image_width - padded_center_crop_size) + 1) // 2
  crop_window = tf.stack([
      offset_height,
      offset_width,
      padded_center_crop_size,
      padded_center_crop_size,
  ])
  image = tf.image.decode_and_crop_jpeg(image_bytes, crop_window, channels=3)
  image = tf.image.resize(
      [image], [_IMAGE_SIZE, _IMAGE_SIZE], method=tf.image.ResizeMethod.BICUBIC
  )[0]
  image = tf.cast(image, tf.int32)

  return image


class Builder(imagenet2012_dataset_builder.Builder):
  """Corrupted ImageNet2012 dataset."""

  BUILDER_CONFIGS = _make_builder_configs()

  def _info(self):
    """Basic information of the dataset.

    Returns:
      tfds.core.DatasetInfo.
    """
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(
                shape=(_IMAGE_SIZE, _IMAGE_SIZE, 3), encoding_format='jpeg'
            ),
            'label': tfds.features.ClassLabel(names_file=names_file),
            'file_name': tfds.features.Text(),  # Eg: 'n15075141_54.JPEG'
        }),
        supervised_keys=('image', 'label'),
        homepage='https://openreview.net/forum?id=HJz6tiCqYm',
    )

  def _split_generators(self, dl_manager):
    """Filter out training split as ImageNet-C is a testing benchmark."""
    splits = super(Builder, self)._split_generators(dl_manager)

    corruptions.FROST_FILENAMES = dl_manager.download(_FROST_FILENAMES)
    return [s for s in splits if s.name != tfds.Split.TRAIN]

  def _generate_examples(
      self, archive, validation_labels=None, labels_exist=None
  ):
    """Generate corrupted imagenet validation data.

    Apply corruptions to the raw images according to self.corruption_type.

    Args:
      archive: an iterator for the raw dataset.
      validation_labels: a dict that maps the file names to imagenet labels.
      labels_exist: unused, to comply with pytype, since overriden method has
        argument too.

    Yields:
      dictionary with the file name, an image file objective, and label of each
      imagenet validation data.
    """
    del labels_exist
    # Get the current random seeds.
    numpy_st0 = np.random.get_state()
    # Set new random seeds.
    np.random.seed(135)
    logging.warning('Overwriting cv2 RNG seed.')
    tfds.core.lazy_imports.cv2.setRNGSeed(357)

    gen_fn = super(Builder, self)._generate_examples
    for key, example in gen_fn(archive, validation_labels):
      with tf.Graph().as_default():
        tf_img = _decode_and_center_crop(example['image'].read())
        image_np = tfds.as_numpy(tf_img)
      example['image'] = self._get_corrupted_example(image_np)

      yield key, example
    # Reset the seeds back to their original values.
    np.random.set_state(numpy_st0)

  def _get_corrupted_example(self, x):
    """Return corrupted images.

    Args:
      x: numpy array, uncorrupted image.

    Returns:
      numpy array, corrupted images.
    """
    corruption_type = self.builder_config.corruption_type
    severity = self.builder_config.severity
    x = np.clip(x, 0, 255)

    return {
        'gaussian_noise': corruptions.gaussian_noise,
        'shot_noise': corruptions.shot_noise,
        'impulse_noise': corruptions.impulse_noise,
        'defocus_blur': corruptions.defocus_blur,
        'glass_blur': corruptions.glass_blur,
        'motion_blur': corruptions.motion_blur,
        'zoom_blur': corruptions.zoom_blur,
        'snow': corruptions.snow,
        'frost': corruptions.frost,
        'fog': corruptions.fog,
        'brightness': corruptions.brightness,
        'contrast': corruptions.contrast,
        'elastic_transform': corruptions.elastic_transform,
        'pixelate': corruptions.pixelate,
        'jpeg_compression': corruptions.jpeg_compression,
        'gaussian_blur': corruptions.gaussian_blur,
        'saturate': corruptions.saturate,
        'spatter': corruptions.spatter,
        'speckle_noise': corruptions.speckle_noise,
    }[corruption_type](x, severity)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import io
from typing import Optional
from unittest import mock

import bs4
from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core.download import downloader
from tensorflow_datasets.core.download import resource as resource_lib
from tensorflow_datasets.core.download import util


class _FakeResponse:

  def __init__(self, url, content, cookies=None, headers=None, status_code=200):
    self.url = url
    self.raw = io.BytesIO(content)
    self.cookies = cookies or {}
    self.headers = headers or {'Content-length': 12345}
    self.status_code = status_code
    # For urllib codepath
    self.read = self.raw.read
    self.text = ''

  def __enter__(self):
    return self

  def __exit__(self, *args):
    return

  def iter_content(self, chunk_size):
    del chunk_size
    for line in self.raw:
      yield line


class DownloaderTest(testing.TestCase):

  def setUp(self):
    super(DownloaderTest, self).setUp()
    self.addCleanup(mock.patch.stopall)
    self.downloader = downloader.get_downloader(10, hashlib.sha256)
    self.tmp_dir = epath.Path(self.tmp_dir)
    self.url = 'http://example.com/foo.tar.gz'
    self.resource = resource_lib.Resource(url=self.url)
    self.path = self.tmp_dir / 'foo.tar.gz'
    self.incomplete_path = self.path.with_suffix(
        self.path.suffix + '.incomplete'
    )
    self.response = b'This \nis an \nawesome\n response!'
    self.resp_checksum = hashlib.sha256(self.response).hexdigest()
    self.cookies = {}
    mock.patch.object(
        downloader.requests.Session,
        'get',
        lambda *a, **kw: _FakeResponse(self.url, self.response, self.cookies),
    ).start()
    self.downloader._pbar_url = mock.MagicMock()
    self.downloader._pbar_dl_size = mock.MagicMock()
    mock.patch.object(
        downloader.urllib.request,
        'urlopen',
        lambda *a, **kw: _FakeResponse(self.url, self.response, self.cookies),
    ).start()

    bs_mock = mock.MagicMock(spec=bs4.BeautifulSoup)
    form_mock = mock.MagicMock()
    form_mock.get.return_value = 'x'
    bs_mock.find.return_value = form_mock
    mock.patch.object(
        bs4, 'BeautifulSoup', autospec=True, return_value=bs_mock
    ).start()

  def test_ok(self):
    promise = self.downloader.download(self.url, self.tmp_dir)
    future = promise.get()
    url_info = future.url_info
    self.assertEqual(self.path, future.path)
    self.assertEqual(url_info.checksum, self.resp_checksum)
    self.assertEqual(self.path.read_bytes(), self.response)
    self.assertFalse(self.incomplete_path.exists())

  def test_drive_no_cookies(self):
    url = 'https://drive.google.com/uc?export=download&id=a1b2bc3'
    promise = self.downloader.download(url, self.tmp_dir)
    future = promise.get()
    url_info = future.url_info
    self.assertEqual(self.path, future.path)
    self.assertEqual(url_info.checksum, self.resp_checksum)
    self.assertEqual(self.path.read_bytes(), self.response)
    self.assertFalse(self.incomplete_path.exists())

  def test_drive(self):
    self.cookies = {'foo': 'bar', 'download_warning_a': 'token', 'a': 'b'}
    self.test_drive_no_cookies()

  def test_http_error(self):
    error = downloader.requests.exceptions.HTTPError('Problem serving file.')
    mock.patch.object(
        downloader.requests.Session, 'get', side_effect=error
    ).start()
    promise = self.downloader.download(self.url, self.tmp_dir)
    with self.assertRaises(downloader.requests.exceptions.HTTPError):
      promise.get()

  def test_bad_http_status(self):
    mock.patch.object(
        downloader.requests.Session,
        'get',
        lambda *a, **kw: _FakeResponse(self.url, b'error', status_code=404),
    ).start()
    promise = self.downloader.download(self.url, self.tmp_dir)
    with self.assertRaises(util.DownloadError):
      promise.get()

  def test_ftp(self):
    url = 'ftp://username:password@example.com/foo.tar.gz'
    promise = self.downloader.download(url, self.tmp_dir)
    future = promise.get()
    url_info = future.url_info
    self.assertEqual(self.path, future.path)
    self.assertEqual(url_info.checksum, self.resp_checksum)
    self.assertEqual(self.path.read_bytes(), self.response)
    self.assertFalse(self.incomplete_path.exists())

  def test_ftp_error(self):
    error = downloader.urllib.error.URLError('Problem serving file.')
    mock.patch.object(
        downloader.urllib.request,
        'urlopen',
        side_effect=error,
    ).start()
    url = 'ftp://example.com/foo.tar.gz'
    promise = self.downloader.download(url, self.tmp_dir)
    with self.assertRaises(downloader.urllib.error.URLError):
      promise.get()


# Gramar examples inspired from: https://tools.ietf.org/html/rfc6266#section-5
_CONTENT_DISPOSITION_FILENAME_PAIRS = [
    ("""attachment; filename=filename.txt""", 'filename.txt'),
    # Should strip space
    ("""attachment; filename=  filename.txt  """, 'filename.txt'),
    ("""attachment; filename=  filename.txt  ;""", 'filename.txt'),
    # If both encoded and ascii are present, only keep encoded
    (
        """attachment; filename="EURO rates"; filename*=utf-8''%e2%82%ac%20rates""",
        'EURO rates',
    ),
    (
        """attachment; filename=EURO rates; filename*=utf-8''%e2%82%ac%20rates""",
        'EURO rates',
    ),
    (
        """attachment; filename=EXAMPLE-Im ößä.dat; filename*=iso-8859-1''EXAMPLE-%20I%27m%20%F6%DF%E4.dat""",
        'EXAMPLE-Im ößä.dat',
    ),
    (
        """attachment;filename="hello.zip";filename*=UTF-8''hello.zip""",
        'hello.zip',
    ),
    (
        """attachment;filename=hello.zip;filename*=UTF-8''hello.zip""",
        'hello.zip',
    ),
    # Should be case insensitive
    ("""INLINE; FILENAME= "an example.html""", 'an example.html'),
    ("""Attachment; filename=example.html""", 'example.html'),
    # Only encoded not supported for now
    ("""attachment; filename*=UTF-8''filename.txt""", None),
    ("""attachment; filename*=iso-8859-1'en'%A3%20rates""", None),
    # Multi-line also supported
    (
        """attachment;
            filename="hello.zip";
            filename*=UTF-8''hello.zip""",
        'hello.zip',
    ),
    ("""attachment;filename*=UTF-8''hello.zip""", None),
    (
        """attachment;
            filename*= UTF-8''%e2%82%ac%20rates.zip""",
        None,
    ),
    # Give only file base name when directory path is given
    (
        'inline;filename=path/to/dir/f-name.png;filename*=UTF-8',
        'f-name.png',
    ),
]


@pytest.mark.parametrize(
    ('content_disposition', 'filename'), _CONTENT_DISPOSITION_FILENAME_PAIRS
)
def test_filename_from_content_disposition(
    content_disposition: str,
    filename: Optional[str],
):
  get_filename = downloader._filename_from_content_disposition
  assert get_filename(content_disposition) == filename


@pytest.mark.parametrize(
    ('content_disposition', 'filename'),
    [
        (
            # Filename should be parsed from the ascii name, not UTF-8
            """attachment;filename="hello.zip";filename*=UTF-8''other.zip""",
            'hello.zip',
        ),
        (
            # If ascii filename can't be parsed, filename parsed from url
            """attachment;filename*=UTF-8''other.zip""",
            'baz.zip',
        ),
        (
            # No headers, filename parsed from url
            None,
            'baz.zip',
        ),
    ],
)
def test_filename_from_headers(
    content_disposition: Optional[str],
    filename: Optional[str],
):
  if content_disposition:
    headers = {
        'content-disposition': content_disposition,
    }
  else:
    headers = None
  resp = _FakeResponse('http://foo.bar/baz.zip', b'content', headers=headers)
  assert downloader._get_filename(resp), filename


@pytest.mark.parametrize(
    ['url', 'filename'],
    [
        (
            'http://test.com/appspot.com/tsvsWithoutLabels%2FAX.tsv?'  # pylint: disable=implicit-str-concat
            'Id=firebase&Expires=2498860800',
            'tsvsWithoutLabels_AX.tsv',  # `%2F` -> `_`
        ),
    ],
)
def test_basename_from_url(url: str, filename: str):
  assert downloader._basename_from_url(url) == filename



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow_datasets as tfds
from tensorflow_datasets.core import visibility

assert visibility._current_available == {
    visibility.DatasetType.TFDS_PUBLIC,
    visibility.DatasetType.COMMUNITY_PUBLIC,
}


def test_visibility():
  # `absl.app` should detect the TFDS script and restrict the visibility
  # to TFDS public by default.
  assert visibility._current_available == {visibility.DatasetType.TFDS_PUBLIC}
  assert 'mnist' in tfds.list_builders()
  assert 'abstract_reasoning' in tfds.list_builders()
  # Remove all visibility and check that the datasets are not available.
  with visibility.set_availables_tmp([]):
    assert 'mnist' not in tfds.list_builders()
    assert 'abstract_reasoning' not in tfds.list_builders()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for croissant_builder."""

from typing import Any, Dict, List, Type
import numpy as np
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core.dataset_builders import croissant_builder
from tensorflow_datasets.core.features import audio_feature
from tensorflow_datasets.core.features import bounding_boxes
from tensorflow_datasets.core.features import bounding_boxes_utils as bb_utils
from tensorflow_datasets.core.features import features_dict
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.features import sequence_feature
from tensorflow_datasets.core.features import tensor_feature
from tensorflow_datasets.core.features import text_feature
from tensorflow_datasets.core.features import video_feature
from tensorflow_datasets.core.utils.lazy_imports_utils import mlcroissant as mlc

FileFormat = file_adapters.FileFormat

DUMMY_DESCRIPTION = "Dummy description."


DUMMY_ENTRIES = [
    {
        "index": i,
        "text": f"Dummy example {i}",
        "split": "train" if i == 0 else "test",
    }
    for i in range(2)
]
DUMMY_ENTRIES_WITH_NONE_VALUES = [
    {"split": "train", "index": 0, "text": "Dummy example 0"},
    {"split": "test", "index": 1, "text": None},
]
DUMMY_ENTRIES_WITH_CONVERTED_NONE_VALUES = [
    {"split": "train", "index": 0, "text": "Dummy example 0"},
    {"split": "test", "index": 1, "text": ""},
]


def _create_mlc_field(
    data_types: mlc.DataType | list[mlc.DataType],
    description: str,
    is_array: bool = False,
    array_shape: str | None = None,
    repeated: bool = False,
    source: mlc.Source | None = None,
    sub_fields: list[mlc.Field] | None = None,
) -> mlc.Field:
  field = mlc.Field(
      data_types=data_types,
      description=description,
      is_array=is_array,
      array_shape=array_shape,
      repeated=repeated,
      sub_fields=sub_fields,
  )
  if source is not None:
    field.source = source
  return field


@pytest.mark.parametrize(
    ["mlc_field", "expected_feature", "int_dtype", "float_dtype"],
    [
        (
            mlc.Field(
                data_types=mlc.DataType.INTEGER, description="Integer feature"
            ),
            np.int64,
            None,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.INT16, description="Int16 feature"
            ),
            np.int16,
            None,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.INTEGER, description="Integer feature"
            ),
            np.int8,
            np.int8,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.FLOAT, description="Float feature"
            ),
            np.float32,
            None,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.FLOAT16, description="Float16 feature"
            ),
            np.float16,
            None,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.FLOAT, description="Float feature"
            ),
            np.float64,
            None,
            np.float64,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.UINT16, description="Uint16 feature"
            ),
            np.uint16,
            None,
            np.uint16,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.BOOL, description="Boolean feature"
            ),
            np.bool_,
            None,
            None,
        ),
    ],
)
def test_simple_datatype_converter(
    mlc_field: mlc.Field,
    expected_feature: type[Any],
    int_dtype: np.dtype | None,
    float_dtype: np.dtype | None,
):
  actual_feature = croissant_builder.datatype_converter(
      mlc_field,
      int_dtype=int_dtype or np.int64,
      float_dtype=float_dtype or np.float32,
  )
  assert actual_feature == expected_feature


def test_datatype_converter_bbox():
  field = _create_mlc_field(
      data_types=mlc.DataType.BOUNDING_BOX,
      description="Bounding box feature",
      source=mlc.Source(format="XYWH"),
  )
  actual_feature = croissant_builder.datatype_converter(field)
  assert isinstance(actual_feature, bounding_boxes.BBoxFeature)
  assert actual_feature.bbox_format == bb_utils.BBoxFormat.XYWH


def test_datatype_converter_bbox_with_invalid_format():
  field = _create_mlc_field(
      data_types=mlc.DataType.BOUNDING_BOX,
      description="Bounding box feature",
      source=mlc.Source(format="InvalidFormat"),
  )
  with pytest.raises(ValueError, match="Unsupported bounding box format"):
    croissant_builder.datatype_converter(field)


@pytest.mark.parametrize(
    ["mlc_field", "feature_type", "subfield_types"],
    [
        (
            mlc.Field(data_types=mlc.DataType.TEXT, description="Text feature"),
            text_feature.Text,
            None,
        ),
        (
            mlc.Field(data_types=mlc.DataType.DATE, description="Date feature"),
            text_feature.Text,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.DATETIME, description="DateTime feature"
            ),
            text_feature.Text,
            None,
        ),
        (
            mlc.Field(data_types=mlc.DataType.TIME, description="Time feature"),
            text_feature.Text,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.IMAGE_OBJECT,
                description="Image feature",
            ),
            image_feature.Image,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.AUDIO_OBJECT,
                description="Audio feature",
            ),
            audio_feature.Audio,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.BOUNDING_BOX,
                description="Bounding box feature",
            ),
            bounding_boxes.BBoxFeature,
            None,
        ),
        (
            mlc.Field(
                data_types=mlc.DataType.VIDEO_OBJECT,
                description="Video feature",
            ),
            video_feature.Video,
            None,
        ),
        (
            mlc.Field(
                id="person",
                data_types=[],
                description="A field with subfields",
                sub_fields=[
                    mlc.Field(id="person/name", data_types=mlc.DataType.TEXT),
                    mlc.Field(id="person/age", data_types=mlc.DataType.INTEGER),
                ],
            ),
            features_dict.FeaturesDict,
            {
                "person/name": text_feature.Text,
                "person/age": tensor_feature.Tensor,
            },
        ),
    ],
)
def test_datatype_converter_complex(
    mlc_field: mlc.Field,
    feature_type: Type[Any],
    subfield_types: Dict[str, Type[Any]] | None,
):
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  expected_description = mlc_field.description
  if isinstance(expected_description, dict):
    expected_description = expected_description.get(
        "en", next(iter(expected_description.values()))
    )
  assert actual_feature.doc.desc == expected_description
  assert isinstance(actual_feature, feature_type)
  if subfield_types is not None:
    for feature_name in actual_feature.keys():
      assert isinstance(
          actual_feature[feature_name], subfield_types[feature_name]
      )


def test_datatype_converter_multilingual_description():
  mlc_field = mlc.Field(
      data_types=mlc.DataType.TEXT,
      description={"en": "English desc", "fr": "Description française"},
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert actual_feature.doc.desc == "English desc"

  mlc_field_no_en = mlc.Field(
      data_types=mlc.DataType.TEXT,
      description={
          "de": "Deutsche Beschreibung",
          "fr": "Description française",
      },
  )
  actual_feature_no_en = croissant_builder.datatype_converter(mlc_field_no_en)
  assert actual_feature_no_en.doc.desc == "Deutsche Beschreibung"


def test_datatype_converter_none():
  field = mlc.Field(
      name="my_field", id="my_field", description="Field with empty data type."
  )
  assert croissant_builder.datatype_converter(field) is None


def test_multidimensional_datatype_converter():
  mlc_field = _create_mlc_field(
      data_types=mlc.DataType.TEXT,
      description="Text feature",
      is_array=True,
      array_shape="2,2",
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert isinstance(actual_feature, tensor_feature.Tensor)
  assert actual_feature.shape == (2, 2)
  assert actual_feature.dtype == np.str_


def test_multidimensional_datatype_converter_image_object():
  mlc_field = _create_mlc_field(
      data_types=mlc.DataType.IMAGE_OBJECT,
      description="Text feature",
      is_array=True,
      array_shape="2,2",
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert isinstance(actual_feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature.feature, image_feature.Image)


def test_multidimensional_datatype_converter_plain_list():
  mlc_field = _create_mlc_field(
      data_types=mlc.DataType.TEXT,
      description="Text feature",
      is_array=True,
      array_shape="-1",
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert isinstance(actual_feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature, text_feature.Text)


def test_multidimensional_datatype_converter_unknown_shape():
  mlc_field = _create_mlc_field(
      data_types=mlc.DataType.TEXT,
      description="Text feature",
      is_array=True,
      array_shape="-1,2",
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert isinstance(actual_feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature.feature, text_feature.Text)


def test_sequence_feature_datatype_converter():
  mlc_field = _create_mlc_field(
      data_types=mlc.DataType.TEXT,
      description="Text feature",
      repeated=True,
  )
  actual_feature = croissant_builder.datatype_converter(mlc_field)
  assert isinstance(actual_feature, sequence_feature.Sequence)
  assert isinstance(actual_feature.feature, text_feature.Text)


@pytest.mark.parametrize(
    ["license_", "expected_license"],
    [
        ("MIT", "MIT"),
        (
            mlc.CreativeWork(
                name="Creative Commons",
                description="Attribution 4.0 International",
                url="https://creativecommons.org/licenses/by/4.0/",
            ),
            (
                "[Creative Commons][Attribution 4.0"
                " International][https://creativecommons.org/licenses/by/4.0/]"
            ),
        ),
        (
            mlc.CreativeWork(
                name="Creative Commons",
            ),
            "[Creative Commons]",
        ),
        (
            mlc.CreativeWork(
                description="Attribution 4.0 International",
            ),
            "[Attribution 4.0 International]",
        ),
        (
            mlc.CreativeWork(
                url="https://creativecommons.org/licenses/by/4.0/",
            ),
            "[https://creativecommons.org/licenses/by/4.0/]",
        ),
        (
            mlc.CreativeWork(),
            "[]",
        ),
    ],
)
def test_extract_license(license_, expected_license):
  actual_license = croissant_builder._extract_license(license_)
  assert actual_license == expected_license


def test_extract_license_with_invalid_input():
  with pytest.raises(
      ValueError, match="^license_ should be mlc.CreativeWork | str"
  ):
    croissant_builder._extract_license(123)


def test_get_license():
  metadata = mlc.Metadata(license=["MIT", "Apache 2.0"])
  actual_license = croissant_builder._get_license(metadata)
  assert actual_license == "MIT, Apache 2.0"


def test_get_license_with_invalid_input():
  with pytest.raises(ValueError, match="metadata should be mlc.Metadata"):
    croissant_builder._get_license(123)


def test_get_license_with_empty_license():
  metadata = mlc.Metadata(license=[])
  assert croissant_builder._get_license(metadata) is None


def test_version_converter(tmp_path):
  with testing.dummy_croissant_file(version="1.0") as croissant_file:
    builder = croissant_builder.CroissantBuilder(
        jsonld=croissant_file,
        file_format=FileFormat.ARRAY_RECORD,
        disable_shuffling=True,
        data_dir=tmp_path,
    )
    assert builder.version == "1.0.0"


@pytest.fixture(name="crs_builder")
def mock_croissant_dataset_builder(
    tmp_path, request
) -> croissant_builder.CroissantBuilder:
  dataset_name = request.param["dataset_name"]
  with testing.dummy_croissant_file(
      dataset_name=dataset_name,
      entries=request.param["entries"],
      split_names=["train", "test"],
  ) as croissant_file:
    builder = croissant_builder.CroissantBuilder(
        jsonld=croissant_file,
        file_format=FileFormat.ARRAY_RECORD,
        disable_shuffling=True,
        data_dir=tmp_path,
    )
    yield builder


@pytest.mark.parametrize(
    "crs_builder",
    [
        {"dataset_name": "DummyDataset", "entries": DUMMY_ENTRIES},
        {
            "dataset_name": "DummyDatasetWithNoneValues",
            "entries": DUMMY_ENTRIES_WITH_NONE_VALUES,
        },
    ],
    indirect=True,
)
def test_croissant_builder(crs_builder):
  assert crs_builder.version == "1.2.0"
  assert (
      crs_builder._info().citation
      == "@article{dummyarticle, title={title}, author={author}, year={2020}}"
  )
  assert crs_builder._info().description == DUMMY_DESCRIPTION
  assert crs_builder._info().homepage == "https://dummy_url"
  assert crs_builder._info().redistribution_info.license == "Public"
  # One `split` and one `jsonl` recordset.
  assert len(crs_builder.metadata.record_sets) == 2
  assert set([rs.id for rs in crs_builder.metadata.record_sets]) == {
      "jsonl",
      "split",
  }
  assert (
      crs_builder.metadata.ctx.conforms_to.value
      == "http://mlcommons.org/croissant/1.0"
  )


@pytest.mark.parametrize(
    "crs_builder,expected_entries",
    [
        (
            {"dataset_name": "DummyDataset", "entries": DUMMY_ENTRIES},
            DUMMY_ENTRIES,
        ),
        (
            {
                "dataset_name": "DummyDatasetWithNoneValues",
                "entries": DUMMY_ENTRIES_WITH_NONE_VALUES,
            },
            DUMMY_ENTRIES_WITH_CONVERTED_NONE_VALUES,
        ),
    ],
    indirect=["crs_builder"],
)
@pytest.mark.parametrize("split_name", ["train", "test"])
def test_download_and_prepare(
    crs_builder: croissant_builder.CroissantBuilder,
    expected_entries: List[Dict[str, Any]],
    split_name: str,
):
  crs_builder.download_and_prepare()
  data_source = crs_builder.as_data_source(split=split_name)
  expected_entries = [
      entry for entry in expected_entries if entry["split"] == split_name
  ]
  assert len(data_source) == 1
  assert len(expected_entries) == 1
  for entry, expected_entry in zip(data_source, expected_entries):
    assert entry["index"] == expected_entry["index"]
    assert entry["text"].decode() == expected_entry["text"]



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to be reused by various imagenet derivatives."""
import os
import tarfile

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_LABELS_FNAME = 'datasets/imagenet2012/labels.txt'

# This file contains the validation labels, in the alphabetic order of
# corresponding image names (and not in the order they have been added to the
# tar file).
_VALIDATION_LABELS_FNAME = 'datasets/imagenet2012/validation_labels.txt'


def label_names_file():
  """Return path to label file."""
  return tfds.core.tfds_path(_LABELS_FNAME)


def get_validation_labels(val_path):
  """Returns labels for validation.

  Args:
    val_path: path to TAR file containing validation images. It is used to
      retrieve the name of pictures and associate them to labels.

  Returns:
    dict, mapping from image name (str) to label (str).
  """
  labels_path = tfds.core.tfds_path(_VALIDATION_LABELS_FNAME)
  with tf.io.gfile.GFile(os.fspath(labels_path)) as labels_f:
    # `splitlines` to remove trailing `\r` in Windows
    labels = labels_f.read().strip().splitlines()
  with tf.io.gfile.GFile(val_path, 'rb') as tar_f_obj:
    tar = tarfile.open(mode='r:', fileobj=tar_f_obj)
    images = sorted(tar.getnames())
  return dict(zip(images, labels))


def generate_examples_validation(archive, labels):
  for fname, fobj in archive:
    record = {
        'file_name': fname,
        'image': fobj,
        'label': labels[fname],
    }
    yield fname, record


def generate_examples_test(archive):
  for fname, fobj in archive:
    record = {
        'file_name': fname,
        'image': fobj,
        'label': -1,
    }
    yield fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Paraphrase Adversaries from Word Scrambling(PAWS) dataset."""

import csv
import os
from typing import Dict

from etils import epath
import tensorflow_datasets.public_api as tfds

_HOMEPAGE_URL = "https://github.com/google-research-datasets/paws"
_LABELED_FINAL = "labeled_final"  # default subset
_LABELED_SWAP = "labeled_swap"
_UNLABELED_FINAL = "unlabeled_final"
_RAW_MAPPING = "raw_and_mapping"
_DOWNLOAD_URLS = {
    _LABELED_FINAL: "https://storage.googleapis.com/paws/english/paws_wiki_labeled_final.tar.gz",
    _LABELED_SWAP: "https://storage.googleapis.com/paws/english/paws_wiki_labeled_swap.tar.gz",
    _UNLABELED_FINAL: "https://storage.googleapis.com/paws/english/paws_wiki_unlabeled_final.tar.gz",
    _RAW_MAPPING: "https://storage.googleapis.com/paws/english/wiki_raw_and_mapping.tar.gz",
}
_EXTRACTED_FOLDERS = {
    _LABELED_FINAL: "final",
    _LABELED_SWAP: "swap",
    _UNLABELED_FINAL: "unlabeled/final",
}
_SUBSET_SPLITS = {
    _LABELED_FINAL: ["train", "dev", "test"],
    _LABELED_SWAP: ["train"],
    _UNLABELED_FINAL: ["train", "dev"],
}
_SPLIT_MAPPINGS = {
    "train": tfds.Split.TRAIN,
    "dev": tfds.Split.VALIDATION,
    "test": tfds.Split.TEST,
}

_CLASS_LABELS = ["different_meaning", "paraphrase"]


class PawsWikiConfig(tfds.core.BuilderConfig):
  """Configuration Class for PAWS Wiki."""

  def __init__(self, tokenized: bool = True, subset: str = "", **kwargs):
    super().__init__(
        name=f"{subset}_{'tokenized' if tokenized else 'raw'}",
        description=f"Subset: {subset} tokenized: {tokenized}",
        **kwargs,
    )
    self.subset = subset
    self.tokenized = tokenized


class Builder(tfds.core.GeneratorBasedBuilder):
  """This is a dataset for paraphrase identification."""

  VERSION = tfds.core.Version("1.1.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial version.",
      "1.1.0": "Adds configs to different subset and support raw text.",
  }
  BUILDER_CONFIGS = [
      PawsWikiConfig(  # pylint: disable=g-complex-comprehension
          subset=subset,
          tokenized=tokenized,
      )
      for tokenized, subset in [
          (True, _LABELED_FINAL),
          (False, _LABELED_FINAL),
          (True, _LABELED_SWAP),
          (False, _LABELED_SWAP),
          (True, _UNLABELED_FINAL),
      ]
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "sentence1": tfds.features.Text(),
            "sentence2": tfds.features.Text(),
            # Label 0: Pair has different meaning, Label 1: Pair is a paraphrase
            "label": tfds.features.ClassLabel(names=_CLASS_LABELS),
        }),
        homepage=_HOMEPAGE_URL,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_paths = dl_manager.download_and_extract(_DOWNLOAD_URLS)
    subset = self.builder_config.subset
    labels_path = os.path.join(dl_paths[subset], _EXTRACTED_FOLDERS[subset])

    if self.builder_config.tokenized:
      mappings_path = ""
      tags2texts = {}
    else:
      mapping_base_dir = os.path.join(
          dl_paths[_RAW_MAPPING], "wiki_raw_and_mapping"
      )
      mappings_path = os.path.join(mapping_base_dir, f"{subset}_mapping")
      swap_path = os.path.join(mapping_base_dir, "input_swap_wiki_50k.tsv")
      backtransl_path = os.path.join(
          mapping_base_dir, "input_backtransl_wiki_with_swap_id.tsv"
      )
      tags2texts = {}
      with epath.Path(swap_path).open() as f:
        reader = csv.DictReader(f, delimiter="\t", quoting=csv.QUOTE_NONE)
        # tsv file format: id  sentence1  sentence2
        for row in reader:
          tags2texts[f"{row['id']}_1"] = row["sentence1"]
          tags2texts[f"{row['id']}_2"] = row["sentence2"]

      with epath.Path(backtransl_path).open() as f:
        reader = csv.DictReader(f, delimiter="\t", quoting=csv.QUOTE_NONE)
        # tsv file format: id  sentence1  sentence2  swap_id
        for row in reader:
          tags2texts[f"{row['id']}_2"] = row["sentence2"]

    return [  # pylint: disable=g-complex-comprehension
        tfds.core.SplitGenerator(
            name=_SPLIT_MAPPINGS[split],
            gen_kwargs={
                "labels_path": os.path.join(labels_path, f"{split}.tsv"),
                "mappings_path": os.path.join(mappings_path, f"{split}.tsv"),
                "tags2texts": tags2texts,
            },
        )
        for split in _SUBSET_SPLITS[subset]
    ]

  def _generate_examples(
      self,
      labels_path: str,
      mappings_path: str,
      tags2texts: Dict[str, str],
  ):
    """Yeilds Examples.

    Args:
      labels_path: The file path for ids, tokenized texts and labels.
      mappings_path: The file path for ids and mapping tags.
      tags2texts: The dictionary of mapping tags to raw texts.

    Yields:
      Generator yielding the next examples
    """
    if tags2texts:
      # uses mapping tags to create dataset with raw texts.
      with epath.Path(labels_path).open() as f:
        # tsv file format: id  sentence1  sentence2  label
        labels = list(csv.DictReader(f, delimiter="\t"))
      with epath.Path(mappings_path).open() as f:
        # tsv file format: id  mapping1  mapping2
        tags = list(csv.DictReader(f, delimiter="\t"))
      if len(labels) != len(tags):
        raise ValueError(
            "Expect same number of labels and mapping tags. "
            f"Got {len(labels)} vs {len(tags)} instead."
        )
      for label_ex, tag_ex in zip(labels, tags):
        if label_ex["id"] != tag_ex["id"]:
          raise ValueError(
              "Expect matched id as key from labels file and mappings file. "
              f"Got {label_ex['id']} vs {tag_ex['id']} instead."
          )
        key = label_ex["id"]
        example = {
            "sentence1": tags2texts[tag_ex["mapping1"]],
            "sentence2": tags2texts[tag_ex["mapping2"]],
            "label": int(label_ex["label"]),
        }
        yield key, example
    else:
      # creates dataset with tokenized texts.
      with epath.Path(labels_path).open() as f:
        reader = csv.DictReader(f, delimiter="\t")
        # tsv file format: id  sentence1  sentence2 label
        for row in reader:
          key = row["id"]
          label_str = (
              "noisy_label"
              if self.builder_config.subset == _UNLABELED_FINAL
              else "label"
          )
          example = {
              "sentence1": row["sentence1"],
              "sentence2": row["sentence2"],
              "label": int(row[label_str]),
          }
          yield key, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Librispeech dataset."""

from __future__ import annotations

import os
from typing import Iterable, List

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = "http://www.openslr.org/12"
_DL_URL = "http://www.openslr.org/resources/12/"
_DL_URLS = {
    "dev_clean": _DL_URL + "dev-clean.tar.gz",
    "dev_other": _DL_URL + "dev-other.tar.gz",
    "test_clean": _DL_URL + "test-clean.tar.gz",
    "test_other": _DL_URL + "test-other.tar.gz",
    "train_clean100": _DL_URL + "train-clean-100.tar.gz",
    "train_clean360": _DL_URL + "train-clean-360.tar.gz",
    "train_other500": _DL_URL + "train-other-500.tar.gz",
}


class LibrispeechConfig(tfds.core.BuilderConfig):
  """Librispeech dataset config."""

  def __init__(self, *args, lazy_decode: bool = False, **kwargs):
    super().__init__(*args, **kwargs)
    self.lazy_decode = lazy_decode


class Builder(tfds.core.GeneratorBasedBuilder):
  """Librispeech dataset."""

  VERSION = tfds.core.Version("2.1.2")
  RELEASE_NOTES = {
      "2.1.2": "Add 'lazy_decode' config.",
      "2.1.1": "Fix speech data type with dtype=tf.int16.",
  }
  MAX_SIMULTANEOUS_DOWNLOADS = 5  # in accordance with http://www.openslr.org
  BUILDER_CONFIGS = [
      LibrispeechConfig(
          name="default",
          description="Default dataset.",
          version=tfds.core.Version("2.1.1"),
      ),
      LibrispeechConfig(
          name="lazy_decode",
          description="Raw audio dataset.",
          lazy_decode=True,
          version=tfds.core.Version("2.1.2"),
      ),
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "speech": tfds.features.Audio(
                sample_rate=16000,
                dtype=np.int16,
                lazy_decode=self.builder_config.lazy_decode,
                file_format="flac",
            ),
            "text": tfds.features.Text(),
            "speaker_id": np.int64,
            "chapter_id": np.int64,
            "id": np.str_,
        }),
        supervised_keys=("speech", "text"),
        homepage=_URL,
        metadata=tfds.core.MetadataDict(
            sample_rate=16000,
        ),
    )

  def _populate_metadata(self, dirs: Iterable[epath.Path]):
    # All dirs contain the same metadata.
    directory = list(dirs)[0]
    self.info.metadata["speakers"] = self._read_metadata_file(
        directory / "LibriSpeech/SPEAKERS.TXT",
        ["speaker_id", "gender", "subset", "minutes", "name"],
    )
    self.info.metadata["chapters"] = self._read_metadata_file(
        directory / "LibriSpeech/CHAPTERS.TXT",
        [
            "chapter_id",
            "speaker_id",
            "minutes",
            "subset",
            "project_id",
            "book_id",
            "chapter_title",
            "project_title",
        ],
    )

  def _read_metadata_file(self, path: epath.Path, field_names: List[str]):
    metadata = {}
    with path.open() as f:
      for line in f:
        if line.startswith(";"):
          continue
        fields = line.split("|", len(field_names))
        metadata[int(fields[0])] = {
            k: v.strip() for k, v in zip(field_names[1:], fields[1:])
        }
    return metadata

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    extracted_dirs = dl_manager.download_and_extract(_DL_URLS)
    self._populate_metadata(extracted_dirs.values())
    splits = {
        split: self._generate_examples(directory)
        for split, directory in extracted_dirs.items()
    }
    return splits

  def _generate_examples(self, directory: epath.Path):
    """Generates examples as dicts."""
    transcripts_glob = os.path.join(directory, "LibriSpeech", "*/*/*/*.txt")
    transcripts_files = tf.io.gfile.glob(transcripts_glob)

    for transcript_file in transcripts_files:
      yield from _generate_librispeech_examples(transcript_file)


def _generate_librispeech_examples(transcript_file: str):
  """Generate examples from a Librispeech transcript file."""
  audio_dir = os.path.dirname(transcript_file)
  with tf.io.gfile.GFile(transcript_file) as f:
    for line in f:
      line = line.strip()
      key, transcript = line.split(" ", 1)
      audio_file = "%s.flac" % key
      speaker_id, chapter_id = [int(el) for el in key.split("-")[:2]]
      example = {
          "id": key,
          "speaker_id": speaker_id,
          "chapter_id": chapter_id,
          "speech": os.path.join(audio_dir, audio_file),
          "text": transcript,
      }
      yield key, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ArrayRecord DataSource base class.

Warning: this is an experimental module. The interface might change in the
future without backwards compatibility.
"""

import dataclasses
from typing import Any, Optional

from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.data_sources import base
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_data_source


@dataclasses.dataclass(repr=False)
class ArrayRecordDataSource(base.BaseDataSource):
  """ArrayRecord DataSource class.

  Warning: this is an experimental class. The interface might change in the
  future without backwards compatibility.

  It acts as a wrapper around `array_record.ArrayRecordDataSource` that can read
  from ArrayRecords. It exposes `__len__` and `__getitem__` to serve as a data
  source.
  """

  dataset_info: dataset_info_lib.DatasetInfo
  split: splits_lib.Split = None
  decoders: Optional[type_utils.TreeDict[decode.partial_decode.DecoderArg]] = (
      None
  )
  deserialize_method: decode.DeserializeMethod = (
      decode.DeserializeMethod.DESERIALIZE_AND_DECODE
  )
  # In order to lazy load array_record, we don't load
  # `array_record_data_source.ArrayRecordDataSource` here.
  data_source: Any = dataclasses.field(init=False)
  length: int = dataclasses.field(init=False)

  def __post_init__(self):
    file_instructions = base.file_instructions(self.dataset_info, self.split)
    self.data_source = array_record_data_source.ArrayRecordDataSource(
        file_instructions
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dolma dataset."""

import gzip
import json

from etils import epath
import tensorflow_datasets.public_api as tfds

_URL = (
    'https://huggingface.co/datasets/allenai/dolma/resolve/main/urls/v1_7.txt'
)


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for Dolma dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Text(),
            'text': tfds.features.Text(),
            'added': tfds.features.Text(),
            'created': tfds.features.Text(),
            'source': tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage='https://github.com/allenai/dolma',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    urls_filepath = dl_manager.download_and_extract(_URL)
    urls = epath.Path(urls_filepath).read_text().splitlines()
    filepaths = dl_manager.download(urls)

    return {
        'train': self._generate_examples(filepaths),
    }

  def _generate_examples(self, filepaths):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam

    def _process_file(file_idx, filepath):
      with epath.Path(filepath).open('rb') as gz_file, gzip.open(gz_file) as f:
        for line_idx, line in enumerate(f):
          row = json.loads(line)
          yield f'{file_idx}_{line_idx}', {
              # Note: there are duplicate ids
              'id': row['id'],
              'text': row['text'],
              'added': str(row.get('added', '')),
              'created': str(row.get('created', '')),
              'source': row.get('source', ''),
          }

    return beam.Create(enumerate(filepaths)) | beam.FlatMapTuple(_process_file)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Physical Interaction QA Dataset."""

import json
import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_PIQA_URL = "https://storage.googleapis.com/ai2-mosaic/public/physicaliqa/physicaliqa-train-dev.zip"


def _read_json(json_path):
  data = []
  with epath.Path(json_path).open() as f:
    for line in f:
      if line:
        data.append(json.loads(line))
  return data


class Builder(tfds.core.GeneratorBasedBuilder):
  """Physical Interaction QA."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "id": tfds.features.Text(),
            "goal": tfds.features.Text(),
            "sol1": tfds.features.Text(),
            "sol2": tfds.features.Text(),
            "label": tfds.features.ClassLabel(num_classes=2),
        }),
        homepage="https://leaderboard.allenai.org/physicaliqa/submissions/get-started",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    extracted_path = dl_manager.download_and_extract(_PIQA_URL)

    # Specify the splits
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "data_path": os.path.join(
                    extracted_path, "physicaliqa-train-dev/train.jsonl"
                ),
                "label_path": os.path.join(
                    extracted_path, "physicaliqa-train-dev/train-labels.lst"
                ),
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "data_path": os.path.join(
                    extracted_path, "physicaliqa-train-dev/dev.jsonl"
                ),
                "label_path": os.path.join(
                    extracted_path, "physicaliqa-train-dev/dev-labels.lst"
                ),
            },
        ),
    ]

  def _generate_examples(self, data_path, label_path):
    """Yields examples."""

    examples = _read_json(data_path)
    labels = _read_json(label_path)

    data = zip(examples, labels)

    for index, (example, label) in enumerate(data):
      example["label"] = label
      yield index, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Asimov 2.0 datasets."""

from tensorflow_datasets.robotics import dataset_importer_builder

ASIMOV_V2_CITATION = """"""

ASIMOV_V2_HOMEPAGE = 'https://asimov-benchmark.github.io/v2/'


class AsimovV2Injuries(dataset_importer_builder.TFDSDatasetImporterBuilder):
  """DatasetBuilder for `asimov_2_injuries` dataset."""

  def get_description(self):
    return 'Situations generated from real hospital injury reports.'

  def get_citation(self):
    return ASIMOV_V2_CITATION

  def get_homepage(self):
    return ASIMOV_V2_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_v2_injuries/0.1.0'


class AsimovV2Videos(dataset_importer_builder.TFDSDatasetImporterBuilder):
  """DatasetBuilder for `asimov_2_videos` dataset."""

  def get_description(self):
    return (
        'Photorealistic videos involving potential physical injury scenarios.'
    )

  def get_citation(self):
    return ASIMOV_V2_CITATION

  def get_homepage(self):
    return ASIMOV_V2_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_v2_videos/0.1.0'


class AsimovV2ConstraintsWithoutRationale(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_v2_constraints_without_rationale` dataset."""

  def get_description(self):
    return (
        'Dataset to evaluate the ability to reason and adhere to'
        ' physical safety constraints imposed by embodiment limitations.'
        ' No rationale for the answer needs to be provided.'
    )

  def get_citation(self):
    return ASIMOV_V2_CITATION

  def get_homepage(self):
    return ASIMOV_V2_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_v2_constraints_without_rationale/0.1.0'


class AsimovV2ConstraintsWithRationale(
    dataset_importer_builder.TFDSDatasetImporterBuilder
):
  """DatasetBuilder for `asimov_v2_constraints_with_rationale` dataset."""

  def get_description(self):
    return (
        'Dataset to evaluate the ability to reason and adhere to'
        ' physical safety constraints imposed by embodiment limitations.'
        ' Rationale for the answer needs to be provided.'
    )

  def get_citation(self):
    return ASIMOV_V2_CITATION

  def get_homepage(self):
    return ASIMOV_V2_HOMEPAGE

  def get_relative_dataset_location(self):
    return 'asimov_v2_constraints_with_rationale/0.1.0'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DataSource base class. See the `Base` class docstring for more context."""

from collections.abc import MappingView, Sequence
import dataclasses
import typing
from typing import Any, Generic, Iterable, Protocol, SupportsIndex, TypeVar

from absl import logging
from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tree

T = TypeVar('T')


@typing.runtime_checkable
class DataSource(Protocol, Generic[T]):
  """Interface for datasources where storage supports efficient random access."""

  def __len__(self) -> int:
    """Returns the total number of records in the data source."""

  def __getitem__(self, key: SupportsIndex) -> T:
    """Returns the value for the given `key`."""

  def __getitems__(self, keys: Iterable[int]) -> T:
    """Returns the value for the given `keys`."""


def file_instructions(
    dataset_info: dataset_info_lib.DatasetInfo,
    split: splits_lib.Split | None = None,
) -> list[shard_utils.FileInstruction]:
  """Retrieves the file instructions from the DatasetInfo."""
  split_infos = dataset_info.splits.values()
  split_dict = splits_lib.SplitDict(split_infos=split_infos)
  return split_dict[split].file_instructions


@dataclasses.dataclass
class BaseDataSource(MappingView, Sequence):
  """Base DataSource to override all dunder methods with the deserialization.

  In order to add a new data source, you can extend BaseDataSource. In the
  __post_init__, you need to define `data_source` which is the underlying data
  source for the file format you define. BaseDataSource will take care of the
  deserialization/decoding.

  Attributes:
    dataset_info: The DatasetInfo of the
    split: The split to load in the data source.
    decoders: Optional decoders for decoding.
    data_source: The underlying data source to initialize in the __post_init__.
    deserialize_method: How to deserialize the bytes that are read before
      returning.
  """

  dataset_info: dataset_info_lib.DatasetInfo
  split: splits_lib.Split | None = None
  decoders: type_utils.TreeDict[decode.partial_decode.DecoderArg] | None = None
  data_source: DataSource[Any] = dataclasses.field(init=False)
  deserialize_method: decode.DeserializeMethod = (
      decode.DeserializeMethod.DESERIALIZE_AND_DECODE
  )

  def _deserialize(self, record: Any) -> Any:
    match self.deserialize_method:
      case decode.DeserializeMethod.RAW_BYTES:
        return record
      case decode.DeserializeMethod.DESERIALIZE_NO_DECODE:
        if file_format := self.dataset_info.file_format:
          return file_format.deserialize(record)
        raise ValueError('No file format set, cannot deserialize bytes!')
      case decode.DeserializeMethod.DESERIALIZE_AND_DECODE:
        if features := self.dataset_info.features:
          return features.deserialize_example_np(record, decoders=self.decoders)  # pylint: disable=attribute-error
        raise ValueError('No features set, cannot decode example!')

  def __getitem__(self, key: SupportsIndex) -> Any:
    record = self.data_source[key.__index__()]
    return self._deserialize(record)

  def __getitems__(self, keys: Sequence[int]) -> Sequence[Any]:
    """Retrieves items by batch.

    This method allows PyTorch to load records by batch, rather than one by one.

    Args:
      keys: a sequence of keys.

    Returns:
      The records associated with the keys.

    Raises:
      IndexError: If the number of retrieved records is incorrect.
    """
    if not keys:
      return []
    records = self.data_source.__getitems__(keys)
    if len(keys) != len(records):
      raise IndexError(
          f'Requested {len(keys)} records but got {len(records)} records.'
          f'{keys=}, {records=}'
      )
    return [self._deserialize(record) for record in records]

  def __repr__(self) -> str:
    decoders_repr = (
        tree.map_structure(type, self.decoders) if self.decoders else None
    )
    return (
        f'{self.__class__.__name__}(name={self.dataset_info.name}, '
        f'split={self.split!r}, '
        f'decoders={decoders_repr})'
    )

  def __len__(self) -> int:
    return self.data_source.__len__()

  def __iter__(self):
    for i in range(self.__len__()):
      yield self[i]



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""simPTE dataset."""

from __future__ import annotations

import csv

import numpy as np
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for simPTE dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  MANUAL_DOWNLOAD_INSTRUCTIONS = """
    Please download training data: sim_pte_train.csv and test data:
    sim_pte_test.csv to ~/tensorflow_datasets/downloads/manual/.
  """

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'y': np.int32,
            'treat': np.int32,
            'X1': np.float32,
            'X2': np.float32,
            'X3': np.float32,
            'X4': np.float32,
            'X5': np.float32,
            'X6': np.float32,
            'X7': np.float32,
            'X8': np.float32,
            'X9': np.float32,
            'X10': np.float32,
            'X11': np.float32,
            'X12': np.float32,
            'X13': np.float32,
            'X14': np.float32,
            'X15': np.float32,
            'X16': np.float32,
            'X17': np.float32,
            'X18': np.float32,
            'X19': np.float32,
            'X20': np.float32,
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=(
            {
                'X1': 'X1',
                'X2': 'X2',
                'X3': 'X3',
                'X4': 'X4',
                'X5': 'X5',
                'X6': 'X6',
                'X7': 'X7',
                'X8': 'X8',
                'X9': 'X9',
                'X10': 'X10',
                'X11': 'X11',
                'X12': 'X12',
                'X13': 'X13',
                'X14': 'X14',
                'X15': 'X15',
                'X16': 'X16',
                'X17': 'X17',
                'X18': 'X18',
                'X19': 'X19',
                'X20': 'X20',
                'treat': 'treat',
            },
            'y',
        ),
        homepage='https://rdrr.io/cran/uplift/man/sim_pte.html',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    train_path = dl_manager.manual_dir / 'sim_pte_train.csv'
    test_path = dl_manager.manual_dir / 'sim_pte_test.csv'

    return {
        'train': self._generate_examples(train_path),
        'test': self._generate_examples(test_path),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    with path.open() as f:
      index = 0
      for row in csv.DictReader(f):
        # And yield (key, feature_dict)
        yield index, {
            'X1': row['X1'],
            'X2': row['X2'],
            'X3': row['X3'],
            'X4': row['X4'],
            'X5': row['X5'],
            'X6': row['X6'],
            'X7': row['X7'],
            'X8': row['X8'],
            'X9': row['X9'],
            'X10': row['X10'],
            'X11': row['X11'],
            'X12': row['X12'],
            'X13': row['X13'],
            'X14': row['X14'],
            'X15': row['X15'],
            'X16': row['X16'],
            'X17': row['X17'],
            'X18': row['X18'],
            'X19': row['X19'],
            'X20': row['X20'],
            'treat': row['treat'],
            'y': row['y'],
        }
        index += 1



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""answer_equivalence dataset."""

from tensorflow_datasets.datasets.answer_equivalence import answer_equivalence_dataset_builder
import tensorflow_datasets.public_api as tfds


class AnswerEquivalenceTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for answer_equivalence dataset."""

  DATASET_CLASS = answer_equivalence_dataset_builder.Builder
  SPLITS = {
      'train': 3,
      'ae_dev': 1,
      'ae_test': 1,
      'dev_xlnet': 1,
      'dev_luke': 1,
      'dev_bidaf': 1,
  }

  DL_EXTRACT_RESULT = {
      'train': 'dummy_data_train.jsonl',
      'ae_dev': 'dummy_data_ae_dev.jsonl',
      'ae_test': 'dummy_data_ae_test.jsonl',
      'dev_xlnet': 'dev_by_system/dummy_data_dev_xlnet.jsonl',
      'dev_luke': 'dev_by_system/dummy_data_dev_luke.jsonl',
      'dev_bidaf': 'dev_by_system/dummy_data_dev_bidaf.jsonl',
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of helper functions to handle dealing with files."""

from __future__ import annotations

import collections
from collections.abc import Iterable, Iterator, Sequence
import concurrent.futures  # pylint: disable=unused-import
import contextlib
import dataclasses
import functools
import os
import random
import re
import string
import time

from absl import logging
from etils import epath
from etils import epy
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import docs
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils import version as version_lib

PathLike = epath.PathLike
ListOrElem = type_utils.ListOrElem
Path = epath.Path

_REGISTERED_DATA_DIRS: set[Path] = set()
_GLOB_CHARS = ['*', '?', '[']

_INFO_FILE_NAMES = [
    constants.FEATURES_FILENAME,
    constants.DATASET_INFO_FILENAME,
]


@dataclasses.dataclass(frozen=True)
class Permissions:
  """Permissions for a file or directory."""

  owner: str | None = None
  group: str | None = None
  mode: int | None = None


@docs.deprecated
def as_path(path: PathLike) -> Path:
  """DEPRECATED. Please use `from etils import epath` with `epath.Path()`."""
  msg = py_utils.dedent(
      """
      `tfds.core.as_path` is deprecated. Pathlib API has been moved to a
      separate module. To migrate, use:

      ```
      from etils import epath
      path = epath.Path('gs://path/to/f.txt')
      ```

      Alternatively `tfds.core.Path` is an alias of `epath.Path`.

      Installation: `pip install etils[epath]`

      """
  )
  logging.warning(msg)
  return Path(path)


def add_data_dir(data_dir: PathLike) -> None:
  """Registers a new default `data_dir` to search for datasets.

  When a `tfds.core.DatasetBuilder` is created with `data_dir=None`, TFDS
  will look in all registered `data_dir` (including the default one) to
  load existing datasets.

  * An error is raised if a dataset can be loaded from more than 1 registered
    data_dir.
  * This only affects reading datasets. Generation always uses the
    `data_dir` kwargs when specified or `tfds.core.constant.DATA_DIR` otherwise.

  Args:
    data_dir: New data_dir to register.
  """
  _REGISTERED_DATA_DIRS.add(Path(data_dir))


def clear_registered_data_dirs() -> None:
  """Clears all registered data_dirs."""
  _REGISTERED_DATA_DIRS.clear()


def _get_incomplete_dir(dir_name: str) -> str:
  """Returns a temporary dir name based on `dir_name`."""
  random_suffix = ''.join(
      random.choice(string.ascii_uppercase + string.digits) for _ in range(6)
  )
  dir_name = Path(dir_name)
  return f'{dir_name.parent}/{constants.INCOMPLETE_PREFIX}{random_suffix}_{dir_name.name}/'


@contextlib.contextmanager
def incomplete_dir(
    dirname: PathLike,
    permissions: Permissions | None = None,
    overwrite: bool = False,
) -> Iterator[str]:
  """Create temporary dir for dirname and rename on exit."""
  dirname = os.fspath(dirname)
  tmp_dir = _get_incomplete_dir(dirname)
  tmp_path = Path(tmp_dir)
  tmp_path.mkdir(parents=True, exist_ok=True)
  try:
    yield tmp_dir
    if overwrite:
      Path(dirname).rmtree(missing_ok=True)
    tmp_path.rename(dirname)
  finally:
    if tmp_path.exists():
      tmp_path.rmtree()


def list_data_dirs(
    given_data_dir: ListOrElem[PathLike] | None = None,
) -> list[PathLike]:
  """Return the list of all `data_dir` to look-up.

  Args:
    given_data_dir: If a `data_dir` is provided, only the explicitly given
      `data_dir` will be returned, otherwise the list of all registered data_dir
      is returned

  Returns:
    The list of all data_dirs to look-up.
  """
  # If the data dir is explicitly given, no need to search everywhere.
  if given_data_dir:
    if isinstance(given_data_dir, list):
      return given_data_dir
    else:
      return [given_data_dir]
  else:
    default_data_dir = Path(constants.DATA_DIR)
    all_data_dirs = _REGISTERED_DATA_DIRS | {default_data_dir}
    return sorted(d.expanduser() for d in all_data_dirs)


def get_dataset_dir(
    data_dir: epath.PathLike,
    builder_name: str,
    config_name: str | None = None,
    version: version_lib.Version | str | None = None,
) -> epath.Path:
  """Returns the data directory for the given dataset."""
  dataset_dir = epath.Path(data_dir) / builder_name
  if config_name:
    dataset_dir /= config_name
  if version:
    dataset_dir /= str(version)
  return dataset_dir


def get_data_dir_and_dataset_dir(
    given_data_dir: PathLike | None,
    builder_name: str,
    config_name: str | None,
    version: version_lib.Version | str | None,
) -> tuple[Path, Path]:
  """Returns the data and dataset directories for the given dataset.

  Args:
    given_data_dir: The data directory to look for the dataset.
    builder_name: The name of the dataset.
    config_name: The config of the dataset.
    version: The version of the dataset.

  Returns:
    data_dir: Root directory containing all datasets, downloads,...
    dataset_dir: Dataset data directory (e.g.
      `<data_dir>/<ds_name>/<config>/<version>`)
  """
  if version is not None:
    version = version_lib.Version(version)

  # If the data dir is explicitly given, no need to search everywhere.
  if given_data_dir is not None:
    given_data_dir = epath.Path(given_data_dir)
    given_dataset_dir = get_dataset_dir(
        data_dir=given_data_dir,
        builder_name=builder_name,
        config_name=config_name,
        version=version,
    )
    return given_data_dir, given_dataset_dir

  # Check whether the dataset is in other data dirs.
  dataset_dir_by_data_dir: dict[Path, Path] = {}
  all_found_versions: set[version_lib.Version] = set()
  for data_dir in list_data_dirs(given_data_dir=None):
    data_dir = Path(data_dir)
    dataset_dir = get_dataset_dir(
        data_dir=data_dir,
        builder_name=builder_name,
        config_name=config_name,
        version=None,
    )
    # Get all versions of the dataset in this dataset dir.
    found_versions = version_lib.list_all_versions(dataset_dir)
    if version in found_versions:
      dataset_dir_by_data_dir[data_dir] = dataset_dir / str(version)
    all_found_versions.update(found_versions)

  if len(dataset_dir_by_data_dir) > 1:
    raise ValueError(
        'Dataset was found in more than one directory:'
        f' {dataset_dir_by_data_dir.values()}. Please resolve the ambiguity'
        ' by explicitly specifying `data_dir=`.'
    )
  elif len(dataset_dir_by_data_dir) == 1:
    # The dataset is found once
    return next(iter(dataset_dir_by_data_dir.items()))

  # No dataset found, use default directory
  default_data_dir = Path(constants.DATA_DIR)
  dataset_dir = get_dataset_dir(
      data_dir=default_data_dir,
      builder_name=builder_name,
      config_name=config_name,
      version=version,
  )
  if all_found_versions:
    logging.warning(
        (
            'Found a different version of the requested dataset'
            ' (given_data_dir=%s, dataset=%s, config=%s, version=%s):\n'
            '%s\nUsing default data dir %s instead.'
        ),
        given_data_dir,
        builder_name,
        config_name,
        version,
        '\n'.join(str(v) for v in sorted(all_found_versions)),
        dataset_dir,
    )
  return default_data_dir, dataset_dir


def _looks_like_a_tfds_file(filename: str) -> bool:
  filename_has_shard_re = re.compile(r'.*\d{5,}-of-\d{5,}.*')
  return bool(filename_has_shard_re.match(filename)) or filename.endswith(
      '.json'
  )


def _find_files_without_glob(
    folder: epath.Path, globs: list[str], file_names: list[str]
) -> Iterator[epath.Path]:
  """Finds files without using globs over nested folders.

  When there is a folder with no permission, then the entire glob fails. This
  function uses `iterdir` to find subfolders.

  Args:
    folder: the folder in which to find files.
    globs: the glob patterns to use to find files.
    file_names: the file names that we are interested in. Used to filter
      results.

  Yields:
    the matching file paths.
  """
  if not folder.is_dir():
    return
  for glob in globs:
    glob_parts = glob.split('/')
    if len(glob_parts) == 1:
      try:
        for file_name in file_names:
          file = folder / file_name
          if file.exists():
            yield file
      except OSError:
        logging.exception('Could not find files in %s', folder)
        continue
    else:
      if glob_parts[0] != '*':
        raise NotImplementedError()
      remaining_glob = '/'.join(glob_parts[1:])
      for subfolder in folder.iterdir():
        if not subfolder.is_dir() or subfolder.name.startswith('.'):
          # Ignore files and folders starting with a dot.
          continue
        yield from _find_files_with_glob(
            subfolder, globs=[remaining_glob], file_names=file_names
        )


def _find_files_with_glob(
    folder: epath.Path,
    globs: list[str],
    file_names: list[str],
) -> Iterator[epath.Path]:
  """Returns files matching any of the given globs and given file names."""
  for glob in globs:
    try:
      found_files = folder.glob(glob)
      for file in found_files:
        if file.name in file_names:
          yield file
    except (
        OSError,
    ):
      # If permission was denied on any subfolder, then the glob fails. Manually
      # iterate through the subfolders instead to be more robust against this.
      yield from _find_files_without_glob(folder, globs, file_names)


def list_dataset_versions(
    dataset_config_dir: epath.PathLike,
) -> list[version_lib.Version]:
  """Returns all dataset versions (sorted ascendingly) found in `dataset_config_dir`.

  Checks whether the version is a valid TFDS version and whether the folder
  contains a dataset_info.json file.

  Arguments:
    dataset_config_dir: the folder that contains version subfolders.
  """
  dataset_config_dir = epath.Path(dataset_config_dir)
  glob = f'*/{constants.DATASET_INFO_FILENAME}'
  found_versions: list[version_lib.Version] = []
  for dataset_info_file in _find_files_with_glob(
      dataset_config_dir,
      globs=[glob],
      file_names=[constants.DATASET_INFO_FILENAME],
  ):
    version_folder = dataset_info_file.parent.name
    if version_lib.Version.is_valid(version_folder):
      found_versions.append(version_lib.Version(version_folder))
  return sorted(found_versions)


def is_valid_variant_dir(
    variant_dir: Path,
    matched_files: set[str] | None = None,
    include_old_tfds_version: bool = False,
) -> bool:
  """Returns whether the variant directory is valid.

  Valid variant directories must:
    - Contain a dataset_info.json file.
    - Contain a features.json file.
    - Have a valid version name.

  Args:
    variant_dir: The variant directory to check.
    matched_files: The files that were matched in the variant directory. If
      None, all json files in the directory are used.
    include_old_tfds_version: include datasets that have been generated with
      TFDS before 4.0.0.
  """
  version = variant_dir.name
  if not version_lib.Version.is_valid(version):
    logging.warning(
        'Variant folder %s has invalid version %s',
        variant_dir,
        version,
    )
    return False

  if matched_files is None:
    matched_files = set(
        matched_path.name for matched_path in variant_dir.glob('*.json')
    )

  if constants.DATASET_INFO_FILENAME not in matched_files:
    logging.warning(
        'Variant folder %s has no %s',
        variant_dir,
        constants.DATASET_INFO_FILENAME,
    )
    return False

  if (
      not include_old_tfds_version
      and constants.FEATURES_FILENAME not in matched_files
  ):
    logging.warning(
        'Variant folder %s has no %s',
        variant_dir,
        constants.FEATURES_FILENAME,
    )
    return False

  return True


def list_dataset_variants(
    dataset_dir: Path,
    namespace: str | None = None,
    include_old_tfds_version: bool = False,
) -> Iterator[naming.DatasetReference]:
  """Yields all variants (config + version) found in `dataset_dir`.

  Args:
    dataset_dir: the folder of the dataset.
    namespace: optional namespace to which this data dir belongs.
    include_old_tfds_version: include datasets that have been generated with
      TFDS before 4.0.0.

  Yields:
    all variants of the given dataset.
  """  # fmt: skip
  data_dir = dataset_dir.parent
  dataset_name = dataset_dir.name
  globs = [
      '*/*/*.json',  # with nested config directory
      '*/*.json',  # without nested config directory
  ]

  # Check files matching the globs and are files we are interested in.
  matched_files_by_variant_dir = collections.defaultdict(set)
  for file in _find_files_with_glob(
      dataset_dir,
      globs=globs,
      file_names=_INFO_FILE_NAMES,
  ):
    matched_files_by_variant_dir[file.parent].add(file.name)

  for variant_dir, matched_files in matched_files_by_variant_dir.items():
    if not is_valid_variant_dir(
        variant_dir=variant_dir,
        matched_files=matched_files,
        include_old_tfds_version=include_old_tfds_version,
    ):
      logging.warning('Skipping invalid variant directory: %s', variant_dir)
      continue

    config_dir = variant_dir.parent
    config = config_dir.name if config_dir != dataset_dir else None
    version = variant_dir.name

    yield naming.DatasetReference(
        namespace=namespace,
        data_dir=data_dir,
        dataset_name=dataset_name,
        config=config,
        version=version,
        info_filenames=matched_files,
    )


def list_datasets_in_data_dir(
    data_dir: epath.PathLike,
    namespace: str | None = None,
    include_configs: bool = True,
    include_versions: bool = True,
    include_old_tfds_version: bool = False,
) -> Iterator[naming.DatasetReference]:
  """Yields references to the datasets found in `data_dir`.

  Only finds datasets that were written to `data_dir`. This means that if
  `data_dir` contains a sub-folder with datasets in them, that these will not be
  found.

  Arguments:
    data_dir: the folder where to look for datasets.
    namespace: optional namespace to which this data dir belongs.
    include_configs: whether to list what configs are available.
    include_versions: whether to list what versions are available.
    include_old_tfds_version: include datasets that have been generated with
      TFDS before 4.0.0.

  Yields:
    references to the datasets found in `data_dir`. The references include the
    data dir.
  """  # fmt: skip
  num_datasets = 0
  num_variants = 0
  for dataset_dir in epath.Path(data_dir).iterdir():
    if not dataset_dir.is_dir():
      continue
    dataset_name = dataset_dir.name
    if not naming.is_valid_dataset_name(dataset_name):
      logging.warning('Invalid dataset name: %s', dataset_name)
      continue
    num_datasets += 1
    if include_configs:
      for variant in list_dataset_variants(
          dataset_dir=dataset_dir,
          namespace=namespace,
          include_old_tfds_version=include_old_tfds_version,
      ):
        num_variants += 1
        if include_versions:
          yield variant
        else:
          yield variant.replace(version=None)
          break
    else:
      num_variants += 1
      yield naming.DatasetReference(
          dataset_name=dataset_name, namespace=namespace, data_dir=data_dir
      )
  logging.info(
      'Found %d datasets and %d variants in %s',
      num_datasets,
      num_variants,
      data_dir,
  )


@functools.lru_cache(maxsize=None)
def makedirs_cached(dirname: epath.PathLike):
  """Creates the given dir with parents and exist is ok.

  Note that this operation is cached, so a dir is only created once. Use this
  with care. If during your session you remove the folder and later you want to
  recreate it, then this method will not do that.

  Arguments:
    dirname: the dir to create.
  """
  epath.Path(dirname).mkdir(parents=True, exist_ok=True)


def expand_glob(path: epath.PathLike) -> Sequence[epath.Path]:
  """Returns all files that match the glob in the given path.

  Warning: If `path` does not contain any wildcards, we do not check whether
  `path` exists and always return `[path]`. In addition, we don't expand the
  `@*` pattern to specify shard wildcards.

  Arguments:
    path: a path that can contain a glob.

  Returns:
    all files that match the given glob.
  """
  path = epath.Path(path).expanduser()
  path_str = os.fspath(path)
  # We don't expand wildcards for number of shards, i.e. paths ending with @*.
  path_str_no_shard_wildcard = re.sub(r'@\*$', '', path_str)
  if not any([char in path_str_no_shard_wildcard for char in _GLOB_CHARS]):
    return [path]
  if not path_str.startswith('/'):
    logging.warning(
        'Can only expand globs for paths starting with a `/`. Got: %s', path
    )
    return [path]
  return list(epath.Path('/').glob(path_str[1:]))


def publish_data(
    from_data_dir: epath.Path,
    to_data_dir: epath.Path,
    overwrite: bool = False,
) -> None:
  """Publishes the data from the given `from_data_dir` to `to_data_dir`.

  Arguments:
    from_data_dir: the folder whose data needs to be published.
    to_data_dir: folder where the data should be published. Should include
      config and version.
    overwrite: whether to overwrite existing data in the `publish_root_dir` if
      it exists.
  """  # fmt: skip
  to_data_dir.mkdir(parents=True, exist_ok=True)
  for filepath in from_data_dir.iterdir():
    filepath.copy(dst=to_data_dir / filepath.name, overwrite=overwrite)


def bulk_rename(
    old_paths: Sequence[epath.PathLike], new_paths: Sequence[epath.PathLike]
) -> None:
  """Renames a sequence of paths in bulk."""
  if len(old_paths) != len(new_paths):
    raise ValueError(
        'old_paths and new_paths must have the same length, but got'
        f' {len(old_paths)} and {len(new_paths)}'
    )
  for old_path, new_path in zip(old_paths, new_paths):
    if old_path == new_path:
      raise ValueError(
          'old_paths and new_paths must not be the same, but got'
          f' {old_path} and {new_path}'
      )
  if not old_paths:
    return
  def _rename(old_and_new_paths: tuple[epath.PathLike, epath.PathLike]):
    old_path, new_path = old_and_new_paths
    epath.Path(old_path).rename(new_path)
  with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
    executor.map(_rename, zip(old_paths, new_paths))


def bulk_delete(paths: Sequence[epath.PathLike]) -> None:
  """Deletes a sequence of paths in bulk."""
  if not paths:
    return
  def _delete(path):
    epath.Path(path).unlink()
  with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
    executor.map(_delete, paths)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for salient_span_wikipedia dataset module."""
from tensorflow_datasets import testing
from tensorflow_datasets.datasets.salient_span_wikipedia import salient_span_wikipedia_dataset_builder


class SalientSpanWikipediaSentencesTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = salient_span_wikipedia_dataset_builder.Builder

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    salient_span_wikipedia_dataset_builder._INPUT_FILE_PATTERN = str(
        cls.dummy_data / 'test_examples.tfrecord.gz'
    )

  BUILDER_CONFIG_NAMES_TO_TEST = ['sentences']
  SPLITS = {'train': 4}


class SalientSpanWikipediaDocumentsTest(SalientSpanWikipediaSentencesTest):
  BUILDER_CONFIG_NAMES_TO_TEST = ['documents']
  SPLITS = {'train': 2}


if __name__ == '__main__':
  testing.test_main()
